// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#ifndef anna_comm_Host_hpp
#define anna_comm_Host_hpp

#include <algorithm>
#include <vector>
#include <map>
#include <string>

#include <anna/core/mt/Mutex.hpp>
#include <anna/core/util/MultiMap.hpp>

namespace anna {

namespace xml {
class Node;
}

namespace comm {

class Communicator;
class Server;
class TransportFactory;
class Device;
class Network;
class ServerAllocator;

/**
   Clase que modela una maquina en la que se ejecutan procesos servidores. Cada maquina contiene
   un nmero indeterminado de procesos servidores (ver Server) a los que enviar peticiones, bien
   directamente, o bien a traves del un sistema de reparto de carga.

   Para crear una nueva maquina hay que invocar a Network::find(const char*) o Network::find(const std::string&).

   \see Server
   \see Service
   \see Network::find
*/
class Host : public Mutex {

  typedef int Port;

  // Requerido por Forte C++
//   struct Comparator;
//   friend struct Comparator;


  struct SortBy {
    static int value(const Server* server) throw();
  };

  std::string a_name;

public:

  typedef std::vector <const Device*> device_container;
  typedef device_container::const_iterator const_device_iterator; /**<Iterador para acceder a los dispositivo de red de la maquina */

  typedef MultiMap <Server, SortBy> server_container;

  typedef server_container::iterator server_iterator; /**<Iterador para acceder a los Servidores asociados a cada uno de los puertos en esta maquina. */
  typedef server_container::const_iterator const_server_iterator; /**<Iterador para acceder a los Servidores asociados a cada uno de los puertos en esta maquina. */


  /**
     Destructor.
  */
  virtual ~Host();

  /**
     Devuelve el nombre logico de esta maquina.

     @return El nombre logico de esta maquina.
  */
  const std::string& getName() const throw() { return a_name; }

  /**
     Crea y registra un nuevo proceso servidor sobre esta maquina (maquina, remotePort) con el nombre indicado.

     @param name Nombre logico del proceso servidor.
     @param remotePort Puerto remoto en el que atiende peticiones este proceso servidor.
     \param autoRecovery Indica si en caso de caida se debe intenrar la recuperacion
     automatica de la conexion.
     \param transportFactory Factoria de protocolos de transporte usada por los ClientSocket asociados a este
     proceso servidor.
     \param ignoreIncomingMessages Indicador de ignorar mensajes entrantes.
     \param doConnect Indicador de connectar al servidor en el momento de crearlo.

     @return La instancia del nuevo proceso servidor
  */
  Server* createServer(const char* name, const int remotePort, const bool autoRecovery, TransportFactory* transportFactory = NULL, const bool ignoreIncomingMessages = false, const bool doConnect = true)
  throw(RuntimeException) {
    return createServer(std::string(name), remotePort, autoRecovery, transportFactory, ignoreIncomingMessages, doConnect);
  }

  /**
     Crea y registra un nuevo proceso servidor sobre esta maquina (maquina, remotePort) con el nombre indicado.

     @param name Nombre logico del proceso servidor.
     @param remotePort Puerto remoto en el que atiende peticiones este proceso servidor.
     \param autoRecovery Indica si en caso de caida se debe intentar la recuperacion
     automatica de la conexion.
     \param transportFactory Factoria de protocolos de transporte usada por los ClientSocket asociados a este
     proceso servidor.
     \param ignoreIncomingMessages Indicador de ignorar mensajes entrantes.
     \param doConnect Indicador de connectar al servidor en el momento de crearlo.

     @return La instancia del nuevo proceso servidor
  */
  Server* createServer(const std::string& name, const int remotePort, const bool autoRecovery, TransportFactory* transportFactory = NULL, const bool ignoreIncomingMessages = false, const bool doConnect = true)
  throw(RuntimeException);

  /**
     Crea y registra un nuevo proceso servidor sobre esta maquina (maquina, remotePort) con el nombre indicado.

     \param serverAllocator Instanciador de Server utilizado para crear la instancia del Server.

     @return La instancia del nuevo proceso servidor
  */
  Server* createServer(const ServerAllocator& serverAllocator) throw(RuntimeException);

  /**
     Devuelve un iterador al comienzo de la lista de servidores asociados a esta maquina.
     \return Un iterador al comienzo de la lista de servidores asociados a esta maquina.
  */
  server_iterator server_begin() throw() { return a_servers.begin(); }

  /**
     Devuelve un iterador al comienzo de la lista de servidores asociados a esta maquina.
     \return Un iterador al comienzo de la lista de servidores asociados a esta maquina.
  */
  const_server_iterator server_begin() const throw() { return a_servers.begin(); }

  /**
     Devuelve un iterador al comienzo de la lista de servidores asociados a esta maquina.
     \return Un iterador al final de la lista de servidores asociados a esta maquina.
  */
  server_iterator server_end() throw() { return a_servers.end(); }

  /**
     Devuelve un iterador al comienzo de la lista de servidores asociados a esta maquina.
     \return Un iterador al final de la lista de servidores asociados a esta maquina.
  */
  const_server_iterator server_end() const throw() { return a_servers.end(); }

  /**
     Devuelve la instancia del servidor sobre el que esta posicionado el iterador recibido
     como parametro.
     \param ii Iterador que debera estar comprendido entre begin y end.
     \return La instancia del servidor sobre el que esta posicionado el iterador recibido
  */
  static Server* server(server_iterator ii) throw() { return server_container::data(ii); }

  /**
     Devuelve la instancia del servidor sobre el que esta posicionado el iterador recibido
     como parametro.
     \param ii Iterador que debera estar comprendido entre begin y end.
     \return La instancia del servidor sobre el que esta posicionado el iterador recibido
  */
  static const Server* server(const_server_iterator ii) throw() { return server_container::data(ii); }


  /**
     Devuelve un iterador al comienzo de la lista de direcciones IPs asociadas a esta maquina.
     \return Un iterador al comienzo de la lista de direcciones IPs asociadas a esta maquina.
  */
  const_device_iterator device_begin() const throw() { return a_devices.begin(); }

  /**
     Devuelve un iterador al final de la lista de direcciones IPs asociadas a esta maquina.
     \return Un iterador al final de la lista de direcciones IPs asociadas a esta maquina.
  */
  const_device_iterator device_end() const throw() { return a_devices.end(); }

  /**
     Devuelve la instancia de la IP sobre el que esta posicionado el iterador recibido
     como parametro.
     \param ii Iterador que debera estar comprendido entre begin y end.
     \return La instancia de la sobre el que esta posicionado el iterador recibido
  */
  static const Device* device(const_device_iterator ii) throw() { return *ii; }

  /**
   * Devuelve la instancia del proceso servidor asociado al puerto recibido como parametro. Si
   * hay ms de una conexin sobre el mismo puerto slo devolver la primera.
   *
   * @param remotePort Puerto remoto en el que atiende peticiones este proceso servidor.
   *
   * @return La instancia del proceso servidor asociado al puerto recibido. Puede ser NULL si no
   * hay ningn proceso servidor asociado con el puerto recibido.
   *
   * @see #createServer
   */
  const Server* find_server(const int remotePort) const throw();

  /**
   * Devuelve la instancia del proceso servidor asociado al puerto recibido como parametro. Si
   * hay ms de una conexin sobre el mismo puerto slo devolver la primera.
   *
   * @param remotePort Puerto remoto en el que atiende peticiones este proceso servidor.
   *
   * @return La instancia del proceso servidor asociado al puerto recibido. Puede ser NULL si no
   * hay ningn proceso servidor asociado con el puerto recibido.
   *
   * @see #createServer
   */
  Server* find_server(const int remotePort) throw();
  /**
     Incorpora un dispositivo de red.
     \param device Dispositivo de red a incorporar.
  */
  void assign(const Device* device) throw(RuntimeException);

  /**
     Devuelve \em true si el dispositivo de red recibido fue asignado a esta maquina o \em
     false en otro caso.
     \param device Dispositivo de red a comprobar.
     \return \em true si el dispositivo de red recibido fue asignado a esta maquina o \em
     false en otro caso.
     \see #assign
  */
  bool contains(const Device* device) const
  throw() {
    const_device_iterator end = device_end();
    return (std::find(device_begin(), end, device) != end);
  }

  /**
     Devuelve una cadena con la informacin referente a esta maquina.
     @return Una cadena con la informacin referente a esta maquina.
  */
  std::string asString() const throw();

  /**
     Devuelve un nodo XML con la informacin referente a este objeto.
     \param parent Nodo XML a partir del cual introducir la informacin.
     \return Un nodo XML con la informacin referente a este objeto.
  */
  xml::Node* asXML(xml::Node* parent) const throw(RuntimeException);


protected:
  /**
     Constructor.
     \param name Nombre logico de esta maquina.
     \warning Cualquier clase heredada que invoque a este constructor debe evitar el acceso
     al metodo #createServer.
  */
  Host(const char* name) : a_name(name) {}

private:
  device_container a_devices;
  server_container a_servers;

  Server* add(Server*, const int remotePort, const bool doConnect) throw();

  friend class Network;
};

}
}

#endif
