// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#ifndef anna_core_mt_Runnable_hpp
#define anna_core_mt_Runnable_hpp

#include <string>

#include <anna/core/functions.hpp>

#include <anna/core/RuntimeException.hpp>
#include <anna/core/mt/Mutex.hpp>

namespace anna {

class Thread;

/**
   Clase generica para implementar clases que deben realizar una unica accion concreta y que puede ser
   lanzada desde un Thread.
*/
class Runnable : public Mutex {
public:
  // Accesores
  /**
     Devuelve el nombre logico de esta instancia.
     @return El nombre logico de esta instancia.
  */
  const std::string& getId() const throw() { return a_id; }

  /**
     Devuelve el valor del indicador de parada.
     \return el valor del indicador de parada.
     \warning La implementacion particular del metodo run deberia comprobar este valor
     periodicamente.
  */
  bool hasRequestedStop() const throw() { return a_requestedStop == true; }

  /**
     Devuelve \em true si la instancia esta en ejecucion dentro de su propio thread o \em false en otro caso.
     \return \em true si la instancia esta en ejecucion dentro de su propio thread o \em false en otro caso.
  */
  bool isRunning() const throw() { return a_isRunning; }

  /**
     Solicita la parada de esta instancia.
  */
  void requestStop() throw(RuntimeException);

  /**
     Devuelve una cadena con la informacion relevante de este objeto.
     \return Una cadena con la informacion relevante de este objeto.
  */
  virtual std::string asString() const
  throw() {
    std::string result("anna::Runnable { Id: ");
    result += a_id;
    result += functions::asText(" | Running: ", a_isRunning);
    result += functions::asText(" | RequestedStop: ", a_requestedStop);
    return result += " }";
  }

protected:
  /**
     Constructor.
  */
  Runnable() : a_id("<none>"), a_requestedStop(false), a_isRunning(false) {;}

  /**
    Constructor.
    @param id Nombre logico de esta instancia.
  */
  Runnable(const std::string& id) : a_id(id), a_requestedStop(false), a_isRunning(false) {;}

  /**
    Constructor.
    @param id Nombre logico de esta clase.
  */
  Runnable(const char* id) : a_id(id), a_requestedStop(false), a_isRunning(false) {;}

  /**
     Establece el nombre logico de esta instancia.
     \param id Nuevo nombre logico de esta instancia.
  */
  void setId(const std::string& id) throw() { a_id = id; }

  /**
     Establece el indicador que informa sobre si esta instancia esta en ejecucion o no.
     \param isRunning Indicador que informa sobre si esta instancia esta en ejecucion.
     \warning Uso interno. Se invoca automaticamente desde Thread.
  */
  void setIsRunning(const bool isRunning) throw() {  a_isRunning = isRunning; }

  /**
     Metodo que se debe reescribir para inicializar el contenido de la clase justo antes de comenzar
     su ejecucion.
  */
  virtual void initialize() throw(RuntimeException) {;}

  /**
     Metodo que debe reescribir para realizar la accion concreta. Por defecto invocara do_action mientras
     que no se invoque la metodo requestStop de la instancia.
  */
  virtual void run() throw(RuntimeException);

  /**
     Metodo que se debe reescribir para finalizar la ejecucion de esta instancia.
     \warning La re-implementacion de este metodo siempre debera invocar a la implementacion
     original que esta re-escribiiendo.
  */
  virtual void terminate() throw() { a_requestedStop = false; }

  /**
     Metodo indicado por Runnable::run en tanto en cuanto no se invoque a requestStop.
  */
  virtual void do_action() throw(RuntimeException) = 0;

private:
  std::string a_id;
  bool a_requestedStop;
  bool a_isRunning;

  friend class Thread;
};

#define anna_complete_runnable(Class) void do_action () throw () {;}

} //namespace anna

#endif

