// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#ifndef anna_core_mt_SafeRecycler_hpp
#define anna_core_mt_SafeRecycler_hpp

#include <stack>
#include <vector>
#include <algorithm>
#include <typeinfo>

#include <anna/core/RuntimeException.hpp>
#include <anna/core/mt/Mutex.hpp>
#include <anna/core/Allocator.hpp>
#include <anna/core/mt/Guard.hpp>
#include <anna/core/util/Recycler.hpp>

namespace anna {

/**
   Mantiene una lista de punteros que puede crecer dinamicamente, no obstante, siempre que sea posible
   intenta reusar punteros creados previamente.

   Establece secciones critidas para acceder a los datos.

   @param T Clase de la que mantener la lista de punteros pre-asignados.
   @param Allocator Clase encargada de reservar la memoria para los objetos T en el momento en que sea necesaria
   una nueva instancia.
*/
template < typename T, typename _Allocator = Allocator <T> >
class SafeRecycler : public Recycler <T, _Allocator>, public Mutex {
public:
  typedef typename Recycler <T, _Allocator>::iterator iterator;
  typedef typename Recycler <T, _Allocator>::const_iterator const_iterator;

  /**
     Constructor.
     \param randomAccess Indicador que permite activar el uso de estructuras de datos adicionales
     Se ha comprobado que si necesitamos tratar en torno a un centenar de instancias
     es ms eficiente no activar las estructuras para acceso directo, para ms objetos resulta
     imprescinble.
  */
  SafeRecycler(const bool randomAccess = false) : Recycler <T, _Allocator> (randomAccess)  {;}

  /**
     Destructor.
  */
  virtual ~SafeRecycler() { ; }

  /**
     Devuelve un puntero de tipo T. Solo crearia una nueva instancia de la clase T si al invocar a este
     metoodo no existe ninguna otra instancia que se pueda reutilizar, en cuyo caso haria una nueva reserva.

     Cada una de las llamadas a este metodo debe tener su correspondiente llamada al metodo  #release cuando
     el puntero deje de ser util.

     @return Un puntero a una instancia de tipo T.
  */
  T* create()
  throw(RuntimeException) {
    std::string name(typeid(*this).name());
    name += "::create";
    Guard guard(this, name.c_str());
    return Recycler <T, _Allocator>::create();
  }

  /**
     Devuelve el iterador que apunta al objeto recibido como parametro.
     \return el iterador que apunta al objeto recibido como parametro.
  */
  iterator find(T* t)
  throw(RuntimeException) {
    std::string name(typeid(*this).name());
    name += "::find";
    Guard guard(this, name.c_str());
    return Recycler <T, _Allocator>::find(t);
  }

  /**
     Libera el puntero recibido como parametro. No se libera fisicamente sino que se deja marcado como
     reusable.

     Si el puntero pasado como parametro no ha sido obtenido mediante el metodo #create los resultados
     no estan definidos.

     @param t Instancia de un puntero de tipo T obtenido a partir del metodo #create.
  */
  void release(T* t)
  throw() {
    if(t == NULL)
      return;

    try {
      std::string name(typeid(*this).name());
      name += "::release (T*)";
      Guard guard(this, name.c_str());
      Recycler <T, _Allocator>::release(t);
    } catch(Exception& ex) {
      ex.trace();
    }
  }

  /**
     Libera el puntero asociado al iterador recibido como parametro.
     \param ii Instancia a liberar.
  */
  void release(iterator& ii) throw() {
    try {
      std::string name(typeid(*this).name());
      name += "::release (iterator)";
      Guard guard(this, name.c_str());
      Recycler <T, _Allocator>::release(ii);
    } catch(Exception& ex) {
      ex.trace();
    }
  }

  /**
     Libera el puntero recibido como parametro. No se libera fisicamente sino que se deja marcado como
     reusable.

     Si el puntero pasado como parametro no ha sido obtenido mediante el metodo #create los resultados
     no estan definidos.

     @param t Instancia de un puntero de tipo T obtenido a partir del metodo #create.
  */
  void release(const T* t) throw() { release(const_cast <T*>(t)); }

  /**
     Marca como disponibles todos los objetos contenidos en memoria.
  */
  void clear()
  throw() {
    std::string name(typeid(*this).name());
    name += "::clear";
    Guard guard(this, name.c_str());
    Recycler <T, _Allocator>::clear();
  }
};

}

#endif

