// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#ifndef anna_core_util_Variable_hpp
#define anna_core_util_Variable_hpp

#include <anna/core/util/defines.hpp>
#include <anna/core/RuntimeException.hpp>
#include <anna/core/define.autoenum.hpp>
#include <anna/core/util/String.hpp>
#include <anna/core/DataBlock.hpp>

namespace anna {

/**
   Recubrimiento de variables de usuario.

   Establece un recubrimiento sobre los tipos de variables soportados por el nucleo
*/
class Variable {
public:
  /**
     Enumeracion con los tipos de datos que podemos asociar a un mensaje interno.
  */
  struct Type {
    enum _v {
      None = -1,
      Unused0,
      String, /**< Cadena de type std::string. */
      Integer, /**< Numero entero de type int. */
      Unused1,
      Integer64, /**< Numero entero de tipo enterno de 64 bits */
      Boolean = 5, /**< Dato de tipo bool. */
      Block = 6, /**< Objeto de tipo DataBlock. */
      Float, /**< Nmero en coma flotante de 32 bits */
      Double, /**< Nmero en coma flotante de 64 bits */
      Custom /** Tipo particular definido por el usuario */
    };
    anna_declare_enum(Type)
  };

  // Constructores
  /**
     Constructor para inicializar una instancia de tipo cadena.

    @param name Nombre logico que recibe este variable.
    @param value Referencia a la variable que estamos recubriendo con esta instancia.
  */
  Variable(const char* name, std::string& value) :
    a_name(name),
    a_isNull(false),
    a_type(Type::String),
    a_isOwner(false) {
    a_value.a_string = &value;
  }


  /**
  * Constructor para inicializar una instancia de type entero.
  *
  * @param name Nombre logico que recibe este variable.
  * @param value Referencia a la variable que estamos recubriendo con esta instancia.
  */
  Variable(const char* name, int& value) :
    a_name(name),
    a_isNull(false),
    a_type(Type::Integer),
    a_isOwner(false) {
    value = 0;
    a_value.a_integer = &value;
  }

  /**
  * Constructor para inicializar una instancia de type long.
  *
  * @param name Nombre logico que recibe este variable.
  * @param value Referencia a la variable que estamos recubriendo con esta instancia.
  */
  Variable(const char* name, S64& value) :
    a_name(name),
    a_isNull(false),
    a_type(Type::Integer64),
    a_isOwner(false) {
    value = 0;
    a_value.a_longInteger = &value;
  }

  /**
  * Constructor para inicializar una instancia de type BOOLEAN.
  *
  * @param name Nombre logico que recibe este variable.
  * @param value Referencia a la variable que estamos recubriendo con esta instancia.
  */
  Variable(const char* name, bool& value) :
    a_name(name),
    a_isNull(false),
    a_type(Type::Boolean),
    a_isOwner(false) {
    value = false;
    a_value.a_boolean = &value;
  }

  /**
  * Constructor para inicializar una instancia de type BLOQUE_MEMORIA.
  *
  * @param name Nombre logico que recibe este variable.
  * @param value Referencia a la variable que estamos recubriendo con esta instancia.
  */
  Variable(const char* name, DataBlock& value) :
    a_name(name),
    a_isNull(false),
    a_type(Type::Block),
    a_isOwner(false) {
    a_value.a_dataBlock = &value;
  }

  /**
  * Constructor para inicializar una instancia de type float.
  *
  * @param name Nombre logico que recibe este variable.
  * @param value Referencia a la variable que estamos recubriendo con esta instancia.
  */
  Variable(const char* name, float& value) :
    a_name(name),
    a_isNull(false),
    a_type(Type::Float),
    a_isOwner(false) {
    value = 0;
    a_value.a_float = &value;
  }

  /**
  * Constructor para inicializar una instancia de type double.
  *
  * @param name Nombre logico que recibe este variable.
  * @param value Referencia a la variable que estamos recubriendo con esta instancia.
  */
  Variable(const char* name, double& value) :
    a_name(name),
    a_isNull(false),
    a_type(Type::Double),
    a_isOwner(false) {
    value = 0;
    a_value.a_double = &value;
  }

  /**
  * Constructor para inicializar una instancia de un tipo definido por el usuario.
  *
  * @param name Nombre logico que recibe este variable.
  * @param value Referencia a la variable que estamos recubriendo con esta instancia.
  */
  Variable(const char* name, void* value) :
    a_name(name),
    a_isNull(false),
    a_type(Type::Custom),
    a_isOwner(false) {
    a_value.a_custom = value;
  }

  /**
     Constructor.
     @param name Nombre logico que recibe esta variable.
     @param type Tipo de dato de esta variable.
  */
  Variable(const char* name, const Type::_v type);

  /**
     Destructor
  */
  virtual ~Variable();

  // Accesores
  /**
     Devuelve el tipo del dato al que recubre esta variable.
     @return El tipo del dato al que recubre esta variable.
  */
  Type::_v getType() const throw() { return a_type; }

  /**
     Devuelve el nombre logico asociado a esta variable.
     \return Nombre logico asociado a esta variable.
  */
  const char* getName() const throw() { return a_name.c_str(); }

  /**
     Devuelve \em false si esta marcado como nulo o \em true en otro caso.
     @return \em false si esta marcado como nulo o \em true en otro caso.
  */
  bool isNull() const throw() { return a_isNull; }

  /**
     Devuelve el valor alfabetico asociado a esta variable.
     @return Si la variable indicada en el contructor de esta instancia puede interpretarse como
     una string devolvera su contenido, en otro caso lanzara una excepcion para indicar el error.
  */
  const char* getStringValue() const throw(RuntimeException);

  /**
     Devuelve el valor numerico asociado a esta variable.
     @return Si la variable indicada en el contructor de esta instancia puede interpretarse como
     un entero de 32 bits devolvera su contenido, en otro caso lanzara una excepcion para indicar el error.
  */
  int getIntegerValue() const throw(RuntimeException);

  /**
     Devuelve el valor numerico asociado a esta variable.
     @return Si la variable indicada en el contructor de esta instancia puede interpretarse como
     un entero de 64 bits devolvera su contenido, en otro caso lanzara una excepcion para indicar el error.
  */
  S64 getInteger64Value() const throw(RuntimeException);

  /**
     Devuelve el valor booleano asociado a esta variable.
     @return Si la variable indicada en el contructor de esta instancia puede interpretarse como
     un booleano y este es distinto de cero devolvera true en otro caso devolvera false.
     indicar el error.
  */
  bool getBooleanValue() const throw(RuntimeException) ;

  /**
     Devuelve el bloque de memoria asociado a esta variable.
     @return Si la variable indicada en el contructor de esta instancia era de type  DataBlock
     devolvera el contenido de esta, en otro caso lanzara una excepcion.
  */
  const DataBlock& getDataBlockValue() const throw(RuntimeException) ;

  /**
     Devuelve el valor numerico asociado a esta variable.
     @return Si la variable indicada en el contructor de esta instancia puede interpretarse como
     un entero en coma flotante devolvera su contenido, en otro caso lanzara una excepcion para indicar el error.
  */
  float getFloatValue() const throw(RuntimeException);

  /**
     Devuelve el valor numerico asociado a esta variable.
     @return Si la variable indicada en el contructor de esta instancia puede interpretarse como
     un entero en coma flotante devolvera su contenido, en otro caso lanzara una excepcion para indicar el error.
  */
  double getDoubleValue() const throw(RuntimeException);

  /**
   * Devuelve el valor de un elemento definido por el usuario.
   */
  void* getCustom() throw(RuntimeException) {
    verifyMatchType(Type::Custom, ANNA_FILE_LOCATION);
    return a_value.a_custom;
  }

  /**
   * Devuelve el valor de un elemento definido por el usuario.
   */
  const void* getCustom() const throw(RuntimeException) {
    verifyMatchType(Type::Custom, ANNA_FILE_LOCATION);
    return a_value.a_custom;
  }

  /**
   * Devuelve el valor de un elemento definido por el usuario.
   */
  void setCustom(void* value) throw(RuntimeException) {
    verifyMatchType(Type::Custom, ANNA_FILE_LOCATION);
    a_isNull = (value == NULL);
    a_value.a_custom = value;
  }

  /**
     Interpreta el valor asociado a esta variable como un valor entero.
     \warning No se realiza ninguna comprobacion semantica.
  */
  int getInteger() const throw() { return *a_value.a_integer; }

  /**
     Interpreta el valor asociado a esta variable como un valor entero largo.
     \warning No se realiza ninguna comprobacion semantica.
  */
  S64 getInteger64() const throw() { return *a_value.a_longInteger; }

  /**
     Interpreta el valor asociado a esta variable como un valor booleano.
     \warning No se realiza ninguna comprobacion semantica.
  */
  bool getBoolean() const throw() { return *a_value.a_boolean; }

  /**
     Interpreta el valor asociado a esta variable como un valor bloque de datos.
     \warning No se realiza ninguna comprobacion semantica.
  */
  const DataBlock& getDataBlock() const throw() { return *a_value.a_dataBlock; }

  /**
     Interpreta el valor asociado a esta variable como un valor en coma flotante de 32 bits
     \warning No se realiza ninguna comprobacion semantica.
  */
  float getFloat() const throw() { return *a_value.a_float; }

  /**
     Interpreta el valor asociado a esta variable como un valor en coma flotante de 64 bits
     \warning No se realiza ninguna comprobacion semantica.
  */
  double getDouble() const throw() { return *a_value.a_double; }

  // Modificadores
  /**
  * Establece el valor de la variable a la que recubre. Si la variable estaba marcada como nula la desmarca.
  * Si el type de dato que deseamos establecer no coincide con el type de dato indicado al crear esta
  * variable devolvera una excepcion de ejecucion indicado la anomalia.
  *
  * @param value Valor que tomara la variable a la que recubre este variable.
  */
  void setValue(const char* value) throw(RuntimeException);

  /**
  * Establece el valor de la variable a la que recubre. Si la variable estaba marcada como nula la desmarca.
  * Si el type de dato que deseamos establecer no coincide con el type de dato indicado al crear esta
  * variable devolvera una excepcion de ejecucion indicado la anomalia.
  *
  * @param value Valor que tomara la variable a la que recubre este variable.
  */
  void setCharPointer(const char* value) throw(RuntimeException);

  /**
  * Establece el valor de la variable a la que recubre. Si la variable estaba marcada como nula la desmarca.
  * Si el type de dato que deseamos establecer no coincide con el type de dato indicado al crear esta
  * variable devolvera una excepcion de ejecucion indicado la anomalia.
  *
  * @param value Valor que tomara la variable a la que recubre este variable.
  */
  void setValue(const int value) throw(RuntimeException);

  /**
  * Establece el valor de la variable a la que recubre. Si la variable estaba marcada como nula la desmarca.
  * Si el type de dato que deseamos establecer no coincide con el type de dato indicado al crear esta
  * variable devolvera una excepcion de ejecucion indicado la anomalia.
  *
  * @param value Valor que tomara la variable a la que recubre este variable.
  */
  void setValue(const S64 value) throw(RuntimeException) ;

  /**
  * Establece el valor de la variable a la que recubre. Si la variable estaba marcada como nula la desmarca.
  * Si el type de dato que deseamos establecer no coincide con el type de dato indicado al crear esta
  * variable devolvera una excepcion de ejecucion indicado la anomalia.
  *
  * @param value Valor que tomara la variable a la que recubre este variable.
  */
  void setValue(const bool value) throw(RuntimeException);

  /**
  * Establece el valor de la variable a la que recubre. Si la variable estaba marcada como nula la desmarca.
  * Si el type de dato que deseamos establecer no coincide con el type de dato indicado al crear esta
  * variable devolvera una excepcion de ejecucion indicado la anomalia.
  *
  * @param value Valor que tomara la variable a la que recubre este variable.
  */
  void setValue(const DataBlock& value) throw(RuntimeException) ;

  /**
  * Establece el valor de la variable a la que recubre. Si la variable estaba marcada como nula la desmarca.
  * Si el type de dato que deseamos establecer no coincide con el type de dato indicado al crear esta
  * variable devolvera una excepcion de ejecucion indicado la anomalia.
  *
  * @param value Valor que tomara la variable a la que recubre este variable.
  */
  void setValue(const float value) throw(RuntimeException) ;

  /**
  * Establece el valor de la variable a la que recubre. Si la variable estaba marcada como nula la desmarca.
  * Si el type de dato que deseamos establecer no coincide con el type de dato indicado al crear esta
  * variable devolvera una excepcion de ejecucion indicado la anomalia.
  *
  * @param value Valor que tomara la variable a la que recubre este variable.
  */
  void setValue(const double value) throw(RuntimeException) ;

  /**
     Marca/Desmarca la variable como nula.
     @param isNull Indica la nueva marca de la variable.
  */
  void setNull(const bool isNull = true) throw() { a_isNull = isNull; }

  /**
     Establece el valor asociado a esta variable como un valor entero.
     \warning No se realiza ninguna comprobacion semantica.
  */
  void setInteger(const int value) throw() { *a_value.a_integer = value; a_isNull = false; }

  /**
     Establece el valor asociado a esta variable como un valor entero largo.
     \warning No se realiza ninguna comprobacion semantica.
  */
  void setLong(const S64 value) throw() { *a_value.a_longInteger = value; a_isNull = false; }

  /**
     Establece el valor asociado a esta variable como un valor booleano.
     \warning No se realiza ninguna comprobacion semantica.
  */
  void setBoolean(const bool value) throw() { *a_value.a_boolean = value; a_isNull = false; }

  /**
     Interpreta el valor asociado a esta variable como un valor bloque de datos.
     \warning No se realiza ninguna comprobacion semantica.
  */
  void setDataBlock(const DataBlock& value) throw() { *a_value.a_dataBlock = value; a_isNull = false; }

  /**
     Interpreta el valor asociado a esta variable como nmero en coma flotante.
     \warning No se realiza ninguna comprobacion semantica.
  */
  void setFloat(const float value) throw() { *a_value.a_float = value; a_isNull = false; }

  /**
     Interpreta el valor asociado a esta variable como nmero en coma flotante.
     \warning No se realiza ninguna comprobacion semantica.
  */
  void setDouble(const double value) throw() { *a_value.a_double = value; a_isNull = false; }

  /**
     @return \em true si la variable recibida tiene el mismo nombre que esta instancia o \em false
     en otro caso.
  */
  bool isEqual(const Variable& right) const throw() { return a_name == right.a_name; }

  /**
     Devuelve una cadena con informacion relevante de esta instancia.
     \return Una cadena con informacion relevante de esta instancia.
  */
  virtual String asString() const throw();

protected:
  /**
     Devuelve la direccion de memoria de la variable C++ a sociada a esta variable logica.
     \return La direccion de memoria de la variable C++ a sociada a esta variable logica.
     \warning La forma de interpretar este valor dependera del tipo de dato (ver #Type) asociado
     Devuelve la referencia a memoria de esta variable.
     \internal
  */
  void* getReference() const throw();

  /**
     Devuelve el area de memoria de la variable C++ a sociada a esta variable logica.
     \return La area de memoria de la variable C++ a sociada a esta variable logica.
     \warning La forma de interpretar este valor dependera del tipo de dato (ver #Type) asociado
     a esta variable.
  */
  void* buffer() const throw();

  /**
     Devuelve la referencia a la variable que controla el indicador de nulo de esta variable.
     \return La referencia a la variable que controla el indicador de nulo de esta variable.
  */
  bool* getNullIndicator() throw() { return &a_isNull; }

private:
  std::string a_name;
  bool a_isNull;
  const Type::_v a_type;
  const bool a_isOwner;

  union {
    std::string* a_string;
    int* a_integer;
    S64* a_longInteger;
    bool* a_boolean;
    DataBlock* a_dataBlock;
    float* a_float;
    double* a_double;
    void* a_custom;
  } a_value;
  struct {
    int integer;
    S64 longInteger;
    bool boolean;
    float theFloat;
    double theDouble;
  } a_aux;

  Variable(const Variable&);
  Variable& operator = (const Variable&);
  void verifyIsNotNull(const char* file, const int lineno) const throw(RuntimeException);
  void verifyMatchSomeType(const Type::_v firstType, const Type::_v secondType, const char* file, const int lineno) const throw(RuntimeException);
  void verifyMatchType(const Type::_v, const char* file, const int lineno) const throw(RuntimeException);

};

}

#endif
