// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#ifndef anna_http_Header_hpp
#define anna_http_Header_hpp

#include <anna/core/RuntimeException.hpp>
#include <anna/core/Allocator.hpp>

namespace anna {

namespace http {

class Token;
class Message;

/**
   Representacion de las cabeceras HTTP.
*/
class Header {
public:
  /**
     Tipos de cabeceras definidos en la RFC 2616.
     \see Header.
  */
  struct Type {
    enum _v {
      None,
      /* Generales */
      CacheControl, Connection, Date, Pragma, Trailer, TransferEncoding, Upgrade, Via,
      Warning,
      /* Request  */
      Accept, AcceptCharset, AcceptEncoding, AcceptLanguage, Authorization, Expect, From,
      Host, IfMatch, IfModifiedSince, IfNoneMatch, IfRange, IfUnmodifiedSince, MaxForwards,
      ProxyAuthorization, Range, Referer, TE, UserAgent,
      /* Entity Header */
      Allow, ContentEncoding, ContentLanguage, ContentLength, ContentLocation, ContentMD5,
      ContentRange, ContentType, Expires, LastModified,
      /* Response Header */
      AcceptRanges, Age, ETAG, Location, ProxyAuthenticate, RetryAfter, Server, Vary,
      WWWAuthenticate,
      /* Uso interno */
      Unknown, End, Begin = 1
    };
  };

  /**
     Categorias de cabeceras definidos en la RFC 2616.
     \see Header.
  */
  struct Category { enum _v { General, Response, Request, Entity, Extension }; };

  /**
     Modos de comparacion aplicados en el metodo #compare.
     \see Header.
  */
  struct Compare { enum _v { Exact = 0, RightTrim = 1, LeftTrim = 2, NoCase = 4, FullMode = 7 }; };

  /**
     Destructor.
  */
  ~Header() { delete a_extensionName; }

  /**
     Devuelve el tipo de la cabecera.
     \return el tipo de la cabecera.
  */
  Type::_v getType() const throw() { return a_type; }

  /**
     Devuelve la categoria de la cabecera.
     \return la categoria de la cabecera.
  */
  Category::_v getCategory() const throw() { return a_category; }

  /**
     Devuelve el puntero nombre de la extension, puede ser NULL.
     \return el puntero nombre de la extension, puede ser NULL.
  */
  const std::string* getExtensionName() const throw() { return a_extensionName; }

  /**
     Devuelve el valor asociado a esta cabecera.
     \return el valor asociado a esta cabecera.
  */
  const std::string& getStringValue() const throw() { return a_value; }

  /**
     Devuelve el valor numerico asociado a esta cabecera.
     \return el valor numerico asociado a esta cabecera.
  */
  const int getIntegerValue() const throw();

  /**
     Establece el valor de esta cabecera.
     \param token Token del que obtendremos el valor.
     \warning Exclusivamente uso interno.
  */
  void setValue(const Token* token) throw();

  /**
     Establece el valor asociado a esta cabecera.
     \param value Valor a establecer.
  */
  void setValue(const std::string& value) throw() { a_value = value; }

  /**
     Establece el valor asociado a esta cabecera.
     \param value Valor a establecer.
  */
  void setValue(const int value) throw();

  /**
     Operador de copia.
     \param other Cabecera de la que copiar.
     \return La instancia de si misma.
  */
  Header& operator = (const Header& other) throw();

  /**
     Compara el contenido actual de esta cabecera con el literal recibido como parametro y
     devuelve  un entero menor, igual o mayor que cero si se encuentra que el contenido
     es,  respectivamente, menor que, igual a (concordante), o mayor que \em str.

     \param str La cadena con la que comparar.
     \param flags Modo de comparacion aplicado.

     \return Un entero menor, igual o mayor que cero si se encuentra que el contenido
     es, respectivamente, menor que, igual a (concordante), o mayor que \em str.
  */
  int compare(const char* str, const int flags = Compare::LeftTrim | Compare::NoCase) const throw();

  /**
     Compara el contenido actual de esta cabecera con el literal recibido como parametro y
     devuelve  \em true o \em false dependiendo de si el contenido coincide con \em str.

     \param str La cadena con la que comparar.
     \param flags Modo de comparacion aplicado.

     \return  \em true o \em false dependiendo de si el contenido coincide con \em str.
  */
  bool match(const char* str, const int flags = Compare::LeftTrim | Compare::NoCase) const throw() {
    return compare(str, flags) == 0;
  }

  /**
     Devuelve una cadena con toda la informacion relevante de este objeto.
     \return una cadena con toda la informacion relevante de este objeto.
  */
  std::string asString() const throw();

  /**
     Interpreta el token recibido como parametro y devuelve el valor de Type con el
     que esta asociado.
     \param token Token obtenido en la fase de analisis.
     \warning Exclusivamente uso interno.
     \return El tipo de cabecera.
  */
  static Type::_v asType(const Token* token) throw();

  /**
     Interpreta el token recibido como parametro y devuelve el nombre de la cabecera
     con la que esta asociado.
     \param type Tipo de cabecera.
     \warning Exclusivamente uso interno.
     \return El literal con el nombre de la cabecera correspondiente al tipo recibido.
  */
  static const char* asLiteral(const Type::_v type) throw();

private:
  Type::_v a_type;
  Category::_v a_category;
  std::string* a_extensionName;
  std::string a_value;
  static const char* st_names [Type::End];

  Header() : a_extensionName(NULL), a_category(Category::Extension) {;}
  Header* initialize(const Type::_v type) throw(RuntimeException);
  Header* initialize(const std::string& name) throw(RuntimeException);
  std::string code() const throw();

  friend class Message;
  friend class Allocator<Header>;
};


}
}

#endif
