// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#include <poll.h>
#include <time.h>
#include <errno.h>
#include <signal.h>
#include <string.h>
#include <sys/ioctl.h>

#include <algorithm>

#include <anna/core/tracing/Logger.hpp>
#include <anna/core/tracing/TraceMethod.hpp>
#include <anna/core/mt/Guard.hpp>
#include <anna/core/util/Average.hpp>
#include <anna/core/mt/ThreadManager.hpp>
#include <anna/core/mt/Thread.hpp>
#include <anna/core/util/Microsecond.hpp>
#include <anna/core/util/defines.hpp>

#include <anna/comm/comm.hpp>
#include <anna/comm/internal/sccs.hpp>
#include <anna/comm/internal/BinderSocket.hpp>
#include <anna/comm/internal/RemoteConnection.hpp>
#include <anna/comm/handler/Manager.hpp>

// ST
#include <anna/comm/internal/Poll.hpp>

// MT
#include <anna/core/mt/ThreadManager.hpp>

#include <anna/xml/Node.hpp>
#include <anna/xml/Attribute.hpp>

using namespace std;
using namespace anna;

// static
const Millisecond comm::Communicator::MinRecoveryTime(1000);
const Millisecond comm::Communicator::DefaultRecoveryTime(5000);
const Millisecond comm::Communicator::MaxRecoveryTime(30000);
const Millisecond comm::Communicator::MinTryingConnectionTime(100);
const Millisecond comm::Communicator::DefaultTryingConnectionTime(200);
const Millisecond comm::Communicator::MaxTryingConnectionTime(1000);
const Millisecond comm::Communicator::DefaultTimeout(10 * 60 * 1000); // 10 minutos.

//static
Millisecond comm::Communicator::st_ReceivingChunkSize(comm::Communicator::DefaultChunkSize);

Communicator::Communicator(const Communicator::WorkMode::_v workMode) :
  Component(getClassName()),
#ifdef _MT
  a_workMode(WorkMode::Single),
#else
  a_workMode(workMode),
#endif
  a_recoveryTime(DefaultRecoveryTime),
  a_requestedStop(false),
  a_status(Status::Available),
  a_isServing(false),
  a_poll(NULL),
  a_threadManager(NULL),
  a_timeout(DefaultTimeout),
  a_mainHandler(NULL),
  a_tryingConnectionTime(DefaultTryingConnectionTime) {
  WHEN_SINGLETHREAD(a_poll = new Poll);
  WHEN_MULTITHREAD(
    a_threadManager = new ThreadManager("comm::Communicator::ThreadManager", ThreadManager::Mode::Unlimit, 0)
  );
  handler::Manager::instantiate().initialize(this);
  a_connectionRecover = new comm::ConnectionRecover(this);
  a_levelOfDenialService = comm::CongestionController::MaxLevel - 1;
  sigignore(SIGPIPE);
  comm::sccs::activate();
}

/*virtual*/
Communicator::~Communicator() {
  delete a_poll;
  delete a_threadManager;
  delete a_connectionRecover;
}

void Communicator::setRecoveryTime(const Millisecond &recoveryTime)
throw(RuntimeException) {
  if(recoveryTime < MinRecoveryTime || recoveryTime > MaxRecoveryTime) {
    string msg("comm::Communicator::setRecoveryTime | ");
    msg += functions::asString("RecoveryTime (%d ms) must be between %d ms and %d ms", recoveryTime.getValue(), MinRecoveryTime.getValue(), MaxRecoveryTime.getValue());
    throw RuntimeException(msg, ANNA_FILE_LOCATION);
  }

  a_recoveryTime = recoveryTime;
}

void Communicator::setTryingConnectionTime(const Millisecond &tryingConnectionTime)
throw(RuntimeException) {
  if(tryingConnectionTime < MinTryingConnectionTime || tryingConnectionTime > MaxTryingConnectionTime) {
    string msg("comm::Communicator::setTryingConnectionTime | ");
    msg += functions::asString("TryingConnectionTime (%d ms) must be between %d ms and %d ms", tryingConnectionTime.getValue(), MinTryingConnectionTime.getValue(), MaxTryingConnectionTime.getValue());
    throw RuntimeException(msg, ANNA_FILE_LOCATION);
  }

  a_tryingConnectionTime = tryingConnectionTime;
}

//static
void Communicator::setReceivingChunkSize(const int receivingChunkSize)
throw(RuntimeException) {
  if(receivingChunkSize < MinReceivingChunkSize || receivingChunkSize > MaxReceivingChunkSize) {
    string msg("comm::Communicator::setReceivingChunkSize | ");
    msg += functions::asString("ReceivingChunkSize (%d bytes) must be between %d bytes and %d bytes", receivingChunkSize, MinReceivingChunkSize, MaxReceivingChunkSize);
    throw RuntimeException(msg, ANNA_FILE_LOCATION);
  }

  st_ReceivingChunkSize = receivingChunkSize;
  CongestionController::instantiate().setMaxPendingBytes(receivingChunkSize);
}

void Communicator::setLevelOfDenialService(const int levelOfDenialService)
throw(RuntimeException) {
  const int min(comm::CongestionController::MaxLevel - 2);
  const int max(comm::CongestionController::MaxLevel);

  if(levelOfDenialService < min || levelOfDenialService > max) {
    string msg("comm::Communicator::setTryingConnectionTime | ");
    msg += functions::asString("LevelOfDenialService %d must be between %d and %d", levelOfDenialService, min, max);
    throw RuntimeException(msg, ANNA_FILE_LOCATION);
  }

  a_levelOfDenialService = levelOfDenialService;
}

void Communicator::attach(ServerSocket* serverSocket)
throw(RuntimeException) {
  if(serverSocket == NULL)
    throw RuntimeException("Cannot attach a NULL comm::ServerSocket", ANNA_FILE_LOCATION);

  Guard guard(this, "comm::Communicator::attach (ServerSocket)");
  Handler* handler = handler::Manager::instantiate().createHandler(serverSocket);
  insert(handler);
  LOGDEBUG(
    string msg("comm::Communicator::attach (ServerSocket) | ");
    msg += handler->asString();
    Logger::debug(msg, ANNA_FILE_LOCATION);
  );

  if(serverSocket->isSharedBind() == true && serverSocket->getBinderSocket() != NULL)
    attach(serverSocket->getBinderSocket());
}

void Communicator::attach(BinderSocket* binderSocket)
throw(RuntimeException) {
  if(binderSocket == NULL)
    throw RuntimeException("Cannot attach a NULL comm::BinderSocket", ANNA_FILE_LOCATION);

  Guard guard(this, "comm::Communicator::attach (BinderSocket)");
  Handler* handler = handler::Manager::instantiate().createHandler(binderSocket);
  insert(handler);
  LOGDEBUG(
    string msg("comm::Communicator::attach (BinderSocket) | ");
    msg += handler->asString();
    Logger::debug(msg, ANNA_FILE_LOCATION);
  );
}

/**
 * Comienza a tratar la conexion que hacen a un comm::ServerSocket de este proceso, desde algun otro proceso.
 *
 * Se invoca desde  comm::handler::ServerSocket::accept [Tx] -> <null>
 */
void Communicator::attach(LocalConnection* localConnection)
throw(RuntimeException) {
  if(localConnection == NULL)
    throw RuntimeException("Cannot attach a NULL comm::LocalConnection", ANNA_FILE_LOCATION);

  /*
   * Obtiene la informacin/bloquea al ClientSocket para asegurar el orden correcto, ya que de otro modo,
   * se podra producir un interbloqueo Communicator & ClientSocket.
   *
   * Recordar que el handler::MetaClientSocket::apply slo bloquea el ClientSocket sobre el que acta y luego
   * y si detecta el cierre,  desbloquea el ClientSocket y bloquea el Communicator. As que para mantener
   * el orden correcto hay que invocar al comm::ClientSocket::getTransport (que establece una SCCS sobre l) antes
   * de bloquea el Communicator.
   */
  ClientSocket* clientSocket = localConnection->getClientSocket();

  if(clientSocket == NULL)
    throw RuntimeException("comm::Communicator::attach (LocalConnection) | ClientSocket can not be NULL", ANNA_FILE_LOCATION);

  // Todava no est corriendo el thread que se encargar de ste comm::ClientSocket => podemos acceder a l sin SSCC.
  const Transport* transport = clientSocket->unsafe_reserveTransport();
  Guard guard(this, "comm::Communicator::attach (LocalConnection)");
  Handler* handler = handler::Manager::instantiate().createHandler(localConnection);
  insert(handler);

  if(((transport == NULL) ? true : transport->enableTimeout()) == true)
    handler->setTimeout(a_timeout);

  WHEN_SINGLETHREAD(

    if(a_workMode == WorkMode::Single && handler->supportTimeout() == true)
    a_timedouts.add(handler)
  );

  LOGDEBUG(
    string msg("comm::Communicator::attach | ");
    msg += handler->asString();
    Logger::debug(msg, ANNA_FILE_LOCATION);
  );

  if(a_workMode == WorkMode::Clone) {
    a_mainHandler = handler;
    app::Application& app = app::functions::getApp().clone();
  }
}

/*
 * Este mtodo slo se invoca desde comm::Server::connect y este mtodo los primero que hace
 * es bloquear el acceso al comunicador, para asegurar el orden correcto de bloqueo a l hora
 * de tratar la desconexin.
 *
 * Por eso no hace falta establecer la seccin crtica que habitualmente se establece en todos
 * los mtodos Communicator::attach.
 */
void Communicator::attach(RemoteConnection* remoteConnection)
throw(RuntimeException) {
  if(remoteConnection == NULL)
    throw RuntimeException("Cannot attach a NULL comm::RemoteConnection", ANNA_FILE_LOCATION);

//   Guard guard (this, "comm::Communicator::attach (RemoteConnection)");
  Handler* handler = handler::Manager::instantiate().createHandler(remoteConnection);
  insert(handler);
  LOGDEBUG(
    string msg("comm::Communicator::attach | ");
    msg += handler->asString();
    Logger::debug(msg, ANNA_FILE_LOCATION);
  );
}

void Communicator::attach(ClientSocket* socket)
throw(RuntimeException) {
  if(socket == NULL)
    throw RuntimeException("Cannot attach a NULL comm::ClientSocket", ANNA_FILE_LOCATION);

  Guard guard(this, "comm::Communicator::attach (ClientSocket)");
  Handler* handler = handler::Manager::instantiate().createHandler(socket);
  insert(handler);
  LOGDEBUG(
    string msg("comm::Communicator::attach | ");
    msg += handler->asString();
    Logger::debug(msg, ANNA_FILE_LOCATION);
  );
}

void Communicator::attach(DatagramSocket* socket)
throw(RuntimeException) {
  if(socket == NULL)
    throw RuntimeException("Cannot attach a NULL comm::DatagramSocket", ANNA_FILE_LOCATION);

  Guard guard(this, "comm::Communicator::attach (DatagramSocket)");
  Handler* handler = handler::Manager::instantiate().createHandler(socket);
  insert(handler);
  LOGDEBUG(
    string msg("comm::Communicator::attach | ");
    msg += handler->asString();
    Logger::debug(msg, ANNA_FILE_LOCATION);
  );
}

void Communicator::attach(Handler* handler)
throw(RuntimeException) {
  if(handler == NULL)
    throw RuntimeException("Cannot attach a NULL comm::Handler", ANNA_FILE_LOCATION);

  if(handler->getType() != Handler::Type::Custom)
    throw RuntimeException("Communicator::attach only accept 'Custom' Handlers", ANNA_FILE_LOCATION);

  Guard guard(this, "comm::Communicator::attach (Handler)");
  insert(handler);
  LOGDEBUG(
    string msg("comm::Communicator::attach (Handler) | ");
    msg += handler->asString();
    Logger::debug(msg, ANNA_FILE_LOCATION);
  );
}

void Communicator::attach(Service* service)
throw(RuntimeException) {
  if(service == NULL)
    throw RuntimeException("Cannot attach a NULL comm::Service", ANNA_FILE_LOCATION);

  if(std::find(service_begin(), service_end(), service) != service_end())
    return;

  service->initialize();
  LOGDEBUG(
    string msg("comm::Communicator::attach | ");
    msg += service->asString();
    Logger::debug(msg, ANNA_FILE_LOCATION);
  );
  a_services.push_back(service);

  if(service->isCritical() == true && service->isAvailable() == false)
    setStatus(Status::Unavailable);
}

void Communicator::insert(Handler* handler)
throw(RuntimeException) {
  handler->initialize();

  if(handler->getfd() < 0) {
    string msg(handler->asString());
    msg += " | Cannot attach a Handler with fd < 0";
    throw RuntimeException(msg, ANNA_FILE_LOCATION);
  }

  const Handler* other = find(handler->getfd());

  if(other != NULL) {
    string msg("commm::Comunicator::insert | New: ");
    msg += handler->asString();
    msg += " | Collision: ";
    msg += other->asString();
    throw RuntimeException(msg, ANNA_FILE_LOCATION);
  }

  a_handlers.add(handler);

  if(handler->supportTimeout())
    handler->beat(anna::functions::hardwareClock());

  WHEN_SINGLETHREAD(a_poll->insert(handler->getfd()));
  WHEN_MULTITHREAD(

    if(a_isServing == true)
    a_threadManager->createThread()->start(*handler);
  );
}

void Communicator::detach(ServerSocket* serverSocket)
throw() {
  if(serverSocket == NULL)
    return;

  comm::BinderSocket* binderSocket = serverSocket->getBinderSocket();
  detach(find(serverSocket->getfd()));

  if(binderSocket != NULL)
    detach(binderSocket);
}

void Communicator::detach(ClientSocket* clientSocket)
throw() {
  if(clientSocket == NULL)
    return;

  detach(find(clientSocket->getfd()));
}

void Communicator::detach(BinderSocket* binderSocket)
throw() {
  if(binderSocket == NULL)
    return;

  detach(find(binderSocket->getfd()));
}

/*
 * Finaliza el trabajo del handler recibido, en este momento NO hay ninguna SSCC establecida.
 * (1) Si se cierra la conexion con el cliente que al que atencia este proceso clonado => debe terminar la ejecucion.
 */
void Communicator::detach(Handler* handler)
throw() {
  if(handler == NULL)
    return;

  Guard guard(this, "comm::Communicator::detach");
  const int fd = handler->getfd();
  LOGDEBUG(
    string msg("comm::Communicator::detach (Handler) | ");
    msg += handler->asString();
    Logger::debug(msg, ANNA_FILE_LOCATION);
  );

  if(anna::functions::supportMultithread() == true)
    handler->requestStop();
  else if(handler->supportTimeout() == true)
    a_timedouts.erase(handler);

  handler->finalize();
  WHEN_SINGLETHREAD(a_poll->erase(fd););

  if(a_handlers.erase(handler) == false) {
    LOGWARNING(
      string msg(functions::asText("Handler: ", fd));
      msg += " | Not found";
      Logger::warning(msg, ANNA_FILE_LOCATION);
    );
  }

  if(a_workMode == WorkMode::Clone && handler == a_mainHandler)                                   // (1)
    requestStop();

  handler::Manager::instantiate().releaseHandler(handler);
}

const Handler* Communicator::getHandler(const ClientSocket& clientSocket)
throw(RuntimeException) {
  Guard guard(this, "comm::Communicator::getHandler");
  Handler* result = find(clientSocket.getfd());

  if(result != NULL) {
    switch(result->getType()) {
    case Handler::Type::ServerSocket:
    case Handler::Type::BinderSocket:
    case Handler::Type::Custom:
      result = NULL;
      break;
    default: break;
    }
  }

  return result;
}

//----------------------------------------------------------------------------------------------
// (1) Compruebo la solicitud de parada antes y despues de procesar el mensaje para acelar
//     la solicitud de la peticion. En otro caso podrian pasar 30 seg desde que se solicita
//     hasta que se acepta.
// (2) La invocacion de este metodo originara que se aniadan y borren fd's a la lista de
// streams pero la forma de recorrer el bucle nos blinda (un poco) de anomalias.
//----------------------------------------------------------------------------------------------
void Communicator::accept()
throw(RuntimeException) {
  LOGMETHOD(TraceMethod traceMethod("comm::Communicator", "accept", ANNA_FILE_LOCATION));

  if(isServing() == true)
    throw RuntimeException("Communicator::accept is already invoked", ANNA_FILE_LOCATION);

  a_requestedStop = false;

  if(handler_size() == 0)
    throw RuntimeException("No socket has been established for sending and/or reception", ANNA_FILE_LOCATION);

  a_isServing = true;
  eventStartup();
  LOGINFORMATION(Logger::write(Logger::Information, Component::asString(), "Polling network", ANNA_FILE_LOCATION));

  try {
    WHEN_SINGLETHREAD(singlethreadedAccept());
    WHEN_MULTITHREAD(multithreadedAccept());
  } catch(RuntimeException& ex) {
    ex.trace();
  }

  a_isServing = false;
  eventShutdown();
}

void Communicator::singlethreadedAccept()
throw(RuntimeException) {
  LOGMETHOD(TraceMethod traceMethod(Logger::Local7, "comm::Communicator", "singlethreadedAccept", ANNA_FILE_LOCATION));
  Handler* handler;
  Microsecond maxTime;
  Microsecond now(anna::functions::hardwareClock());
  int fd;
  a_poll->setTimeout((a_connectionRecover->isRunning() == true) ? a_recoveryTime : (Millisecond)5000);
  maxTime = a_timeout;
  maxTime += now;

  while(a_requestedStop == false) {
    if(a_connectionRecover->isRunning() == true) {
      a_connectionRecover->tryRecover();
      a_poll->setTimeout((a_connectionRecover->isRunning() == true) ? a_recoveryTime : (Millisecond)5000);
    }

    a_poll->waitMessage();

    if(a_timedouts.size() > 0)
      now = anna::functions::hardwareClock();

    while((fd = a_poll->fetch()) != -1) {
      if((handler = find(fd)) == NULL)
        continue;

      try {
        if(handler->supportTimeout())
          handler->beat(now);

        handler->apply();
      } catch(RuntimeException& ex) {
        ex.trace();
      }
    }

    if(a_pendingClose == true) {
      a_pendingClose = false;

      // @Eduardo (multiconnection to same address/port); On st, is considered to have one socket pending to be closed
      for(handler_iterator ii = handler_begin(), maxii = handler_end(); ii != maxii; ii ++)
//            if (Communicator::handler (ii)->testClose () == true)
//               break;
        Communicator::handler(ii)->testClose();
    }

    now = anna::functions::hardwareClock();

    if(a_timedouts.size() == 0)
      continue;

    if(now > maxTime) {
      handler_iterator ii = a_timedouts.begin();

      while(ii != a_timedouts.end()) {
        handler = Communicator::handler(ii);

        if(handler->isTimeout(now) == true) {
          LOGWARNING(
            string msg(handler->asString());
            msg += " | Closed due to inactivity";
            Logger::warning(msg, ANNA_FILE_LOCATION);
          );
          detach(handler);
          ii = a_timedouts.begin();
        } else
          ii ++;
      }

      maxTime = a_timeout;
      maxTime += now;
    }
  }
}

void Communicator::multithreadedAccept()
throw(RuntimeException) {
  LOGMETHOD(TraceMethod traceMethod(Logger::Local7, "comm::Communicator", "multithreadedAccept", ANNA_FILE_LOCATION));
  {
    Guard guard(this, "comm::Communicator::multithreadedAccept");
    Handler* handler;

    for(handler_iterator ii = handler_begin(), maxii = handler_end(); ii != maxii; ii ++) {
      handler = Communicator::handler(ii);
      LOGDEBUG(
        string msg("Starting | ");
        msg += handler->asString();
        Logger::debug(msg, ANNA_FILE_LOCATION)
      );
      a_threadManager->createThread()->start(*handler);
    }
  }
  Millisecond delay(500);

  while(a_requestedStop == false) {
    anna::functions::sleep(delay);

    if(a_requestedStop == true)
      break;

    if(a_connectionRecover->isRunning() == false)
      continue;

    {
      Guard guard(this, "comm::Communicator::multithreadedAccept (ConnectionRecover)");
      a_connectionRecover->tryRecover();
    }
  }
}

void Communicator::requestStop()
throw() {
  if(a_requestedStop == true)
    return;

  Guard guard(this, "comm::Communicator::requestStop");

  if(a_requestedStop == true)
    return;

  a_requestedStop = true;

  try  {
    for(handler_iterator ii = handler_begin(), maxii = handler_end(); ii != maxii; ii ++)
      handler(ii)->requestStop();
  } catch(RuntimeException& ex) {
    ex.trace();
  }

  LOGWARNING(
    string msg("comm::Communicator::requestStop | ");
    msg += asString();
    Logger::warning(msg, ANNA_FILE_LOCATION);
  );
}

bool Communicator::isUsable(const ClientSocket* clientSocket)
throw() {
  if(clientSocket == NULL)
    return false;

  Guard guard(this, "comm::Communicator::isUsable");
  Handler* handler = find(clientSocket->getfd());

  if(handler == NULL)
    return false;

  const Handler::Type::_v type = handler->getType();
  return (type == Handler::Type::LocalConnection || type == Handler::Type::RemoteConnection);
}

void Communicator::setStatus(const Status& status)
throw() {
  Guard guard(this, "comm::Communicator::setStatus");

  if(a_status != status)
    a_status = status;

  LOGINFORMATION(
    string msg("comm::Communicator::setStatus | ");
    msg += a_status.asString();
    Logger::information(msg, ANNA_FILE_LOCATION);
  );
}

void Communicator::eventBreakAddress(const in_addr_t& address)
throw() {
  Device* device = Network::instantiate().find(address);

  if(device->getStatus() == Device::Status::Down)
    return;

  LOGWARNING(
    string msg("comm::Communicator::eventBreakAddress | ");
    msg += device->asString();
    Logger::warning(msg, ANNA_FILE_LOCATION);
  );
  Guard guard(this, "comm::Communicator::eventBreakAddress");
  device->setStatus(Device::Status::Down);
  /**
   * Trabaja sobre una copia para no perder la referencia cuando se elimine un miembro de la lista original
   *
   * En la lista a borrar slo mete los handler::ServerSocket y los handler::RemoteConnection, ya que los handler::LocalConnection
   * ser liberados recursivamente cuando liberemos los primeros.
   */
  typedef vector <comm::Handler*> work_container;
  typedef work_container::iterator work_iterator;
  work_container ww;

  for(handler_iterator ii = handler_begin(), maxii = handler_end(); ii != maxii; ii ++) {
    comm::Handler* handler = comm::Communicator::handler(ii);

    if(dynamic_cast <comm::handler::ServerSocket*>(handler) != NULL) {
      ww.push_back(handler);
    } else if(dynamic_cast <comm::handler::RemoteConnection*>(handler) != NULL) {
      ww.push_back(handler);
    }
  }

  for(work_iterator ii = ww.begin(), maxii = ww.end(); ii != maxii; ii ++) {
    LOGDEBUG(
      std::string msg("Communicator::eventBreakAddress | ");
      msg = (*ii)->asString();
      Logger::debug(msg, ANNA_FILE_LOCATION)
    );
    (*ii)->breakAddress(address);
  }
}

void Communicator::eventRecoverAddress(const in_addr_t& address)
throw() {
  Device* device = Network::instantiate().find(address);

  if(device->getStatus() == Device::Status::Up)
    return;

  LOGWARNING(
    string msg("comm::Communicator::eventRecoverAddress | ");
    msg += device->asString();
    Logger::warning(msg, ANNA_FILE_LOCATION);
  );
  Guard guard(this, "comm::Communicator::eventRecoverAddress");
  device->setStatus(Device::Status::Up);
  Handlers backup(a_handlers);

  for(handler_iterator ii = backup.begin(), maxii = backup.end(); ii != maxii; ii ++)
    handler(ii)->recoverAddress(address);
}

bool Communicator::eventAcceptConnection(const ClientSocket& clientSocket)
throw(RuntimeException) {
  LOGMETHOD(TraceMethod traceMethod(Logger::Local7, "comm::Communicator", "eventAcceptConnection", ANNA_FILE_LOCATION));

  if(a_requestedStop == true) {
    LOGWARNING(
      string msg(Component::asString());
      msg += " | ";
      msg += clientSocket.asString();
      msg += " | Connection rejected due to stop request";
      Logger::warning(msg, ANNA_FILE_LOCATION);
    );
    return false;
  }

  CongestionController::Workload ww = CongestionController::instantiate().getAccumulatedWorkload();

  if(CongestionController::getLevel(ww) >= a_levelOfDenialService) {
    LOGWARNING(
      string msg("comm::Communicator::eventAcceptConnection | Level: ");
      msg += functions::asString(CongestionController::getLevel(ww));
      msg += functions::asString(" | Load: %d%% ", CongestionController::getLoad(ww));
      msg += " | Result: false";
      Logger::warning(msg, ANNA_FILE_LOCATION);
    );
    return false;
  }

  return true;
}

//--------------------------------------------------------------------------------------------------------
// (1) Alguno de los comm::Server asociados al servicio puede haber pasado a activo, pero nos
//     interesa como estaba el servicio antes de esta activacion.
// (2) Si el servicio es "No critico" => no afecta al estado del proceso.
// (3) Solo si todos los servicios "Criticos" estan disponibles pasara a estar "Activo".
//--------------------------------------------------------------------------------------------------------
void Communicator::eventCreateConnection(const Server* server)
throw() {
  if(server == NULL)
    return;

  LOGMETHOD(TraceMethod traceMethod(Logger::Local7, "comm::Communicator", "eventCreateConnection", ANNA_FILE_LOCATION));
  LOGNOTICE(
    string msg("comm::Communicator::eventCreateConnection | ");
    msg += server->asString();
    Logger::notice(msg, ANNA_FILE_LOCATION);
  );
  bool recoverService = false;

  for(Server::const_iterator ii = server->begin(), maxii = server->end(); ii != maxii; ii ++) {
    Service* service = const_cast <Service*>(Server::service(ii));

    if(service->wasAvailable() == true) {                                             // (1)
      service->recover(server);
      continue;
    }

    service->recover(server);
    eventCreateConnection(service);
    recoverService = true;
  }

  if(recoverService == false || a_status == Status::Available)
    return;

  Guard guard(this, "comm::Communicator::eventCreateConnection");
  Status status(Status::Available);

  for(const_service_iterator ii = service_begin(), maxii = service_end(); ii != maxii; ii ++) {
    const Service* service = Communicator::service(ii);

    if(service->isCritical() == false)                                               // (2)
      continue;

    if(service->isAvailable() == false) {
      status = Status::Unavailable;
      break;
    }
  }

  setStatus(status);                                                                  // (3)
}

void Communicator::eventCreateConnection(const Service* service)
throw() {
  if(service == NULL)
    return;

  LOGNOTICE(
    string msg("comm::Communicator::eventCreateConnection | ");
    msg += service->asString();
    Logger::notice(msg, ANNA_FILE_LOCATION);
  );
}

/*
 * Se invoca desde handler::RemoteConnection:[Tx] -> Communicator
 */
void Communicator::eventBreakConnection(const Server* server)
throw() {
  if(server == NULL)
    return;

  LOGMETHOD(TraceMethod traceMethod(Logger::Local7, "comm::Communicator", "eventBreakConnection (server)", ANNA_FILE_LOCATION));
  LOGWARNING(
    string msg("comm::Communicator::eventBreakConnection | ");
    msg += server->asString();
    Logger::warning(msg, ANNA_FILE_LOCATION);
  );
  //Guard guard (this, "comm::Communicator::eventBreakConnection");
  Status status(a_status);

  for(Server::const_iterator ii = server->begin(), maxii = server->end(); ii != maxii; ii ++) {
    Service* service = const_cast <Service*>(Server::service(ii));
    service->fault(server);

    if(service->isAvailable() == true)
      continue;

    eventBreakConnection(service);

    if(status == Status::Available && service->isCritical() == true)
      status = Status::Unavailable;
  }

  setStatus(status);
}

void Communicator::eventBreakConnection(const Service* service)
throw() {
  if(service == NULL)
    return;

  LOGMETHOD(TraceMethod traceMethod(Logger::Local7, "comm::Communicator", "eventBreakConnection (service)", ANNA_FILE_LOCATION));
  LOGWARNING(
    string msg("comm::Communicator::eventBreakConnection | ");
    msg += service->asString();
    Logger::warning(msg, ANNA_FILE_LOCATION);
  );
}

void Communicator::eventShutdown()
throw() {
  LOGWARNING(
    string msg("comm::Communicator::eventShutdown | ");
    msg += asString();
    Logger::warning(msg, ANNA_FILE_LOCATION);
  );
  setStatus(Status::Unavailable);
#ifndef _MT
  Handlers backup(a_handlers);

  for(handler_iterator ii = backup.begin(), maxii = backup.end(); ii != maxii; ii ++)
    detach(handler(ii));

#else

  try {
    a_threadManager->join();
  } catch(RuntimeException& ex) {
    ex.trace();
  }

#endif
}

std::string Communicator::asString() const
throw() {
  string result("comm::Communicator { ");
  result += Component::asString();
  result += " | RequestedStop: ";
  result += anna::functions::asString(a_requestedStop);
  result += " | ";
  result += a_status.asString();
  result += anna::functions::asString(" | Handlers: %d", a_handlers.size());
  result += anna::functions::asString(" | RecoveryTime: %d ms", a_recoveryTime.getValue());
  return result += " }";
}

xml::Node* Communicator::asXML(xml::Node* parent) const
throw() {
  parent = app::Component::asXML(parent);
  xml::Node* result = parent->createChild("comm.Communicator");
  result->createAttribute("RequestedStop", anna::functions::asString(a_requestedStop));
  result->createAttribute("RecoveryTime", a_recoveryTime);
  result->createAttribute("TryingConnectionTime", a_tryingConnectionTime);
  result->createAttribute("Timeout", a_timeout);
  result->createAttribute("Status", a_status.asString());
  result->createAttribute("Mode", (a_workMode == WorkMode::Single) ? "Single" : "Clone");
  result->createAttribute("LevelOfDenialService", a_levelOfDenialService);
  Network::instantiate().asXML(result);
  xml::Node* node = result->createChild("comm.Handlers");

  for(const_handler_iterator ii = handler_begin(), maxii = handler_end(); ii != maxii; ii ++)
    handler(ii)->asXML(node);

  node = result->createChild("comm.Services");

  for(const_service_iterator ii = service_begin(), maxii = service_end(); ii != maxii; ii ++)
    service(ii)->asXML(node);

  a_connectionRecover->asXML(result);
  CongestionController& congestionController = CongestionController::instantiate();
  congestionController.asXML(result);
  return result;
}

/*
 * Se invoca desde app::Application::clone -> app::Component::do_cloneParent (ojo EN EL PROCESO ORIGINAL).
 * Ofrece la posiblidad de que el componente del proceso original liberen los recursos que no va
 * a usar.
 *
 * Cada vez que se clona tiene que sacar de su comprobacion las conexiones que ha sufrido. Ya que de otro
 * modo podria estar tratantado contestaciones a peticiones realizadas desde los procesos hijos => estos
 * nunca recibirian las respuestas a sus peticiones.
 *
 * Estas conexiones no se pueden cerrar porque deben mantenerse abiertas para que los procesos hijos las
 * hereden, solo es que este proceso no debe atenderlas.
 *
 * El codigo del servidor (el proceso original) no hace nada con ese Socket, lo cierra porque ya lo ha duplicado
 * el proceso hijo y sigue atendiendo peticiones de conexion.
 */
void Communicator::do_cloneParent()
throw(RuntimeException) {
  LOGMETHOD(TraceMethod traceMethod("comm::Communicator", "do_cloneParent", ANNA_FILE_LOCATION));
  Handler* handler;
  Handler::Type::_v type;

  for(handler_iterator ii = handler_begin(), maxii = handler_end(); ii != maxii; ii ++) {
    handler = Communicator::handler(ii);
    type = handler->getType();

    if(type == Handler::Type::RemoteConnection || type == Handler::Type::ClientSocket)
      a_poll->erase(handler->getfd());
  }

  try {
    handler = a_mainHandler;
    a_mainHandler = NULL;
    detach(handler);
  } catch(RuntimeException& ex) {
    ex.trace();
  }
}

/*
 * Se invoca desde app::Application::clone -> app::Component::do_cloneChild (ojo EN EL NUEVO PROCESO).
 * Ofrece la posiblidad de que los componentes que acaban de ser duplicados liberen los recursos que no van
 * a usar, ya que la copia solo se dedicara a tratar los mensajes que entren por su conexion asociada y
 * las respuestas a las peticiones originadas en el tratamiento del mismo.
 *
 * (1) Recordar que el handler que tiene asociado el clon esta registrado para evitar su cierre.
 * (3) Limpia la lista porque de otro modo si el proceso clon hiciera nuevas conexiones podria encontrar
 *     que el fd ya esta guardado en la lista. No se puede invocar directamente al detach, porque este
 *     metodo modifica la lista a_handlers, y nos haria perder la cuenta del iterator.
 * (4) Realiza un duplicado fisico de las conexiones realizadas por el proceso original, elimina el fd de
 *     la lista a comprobar, este fd sera cerrado por el metodo Handler::clone.
 * (5) Registra el Handler por el que ha sido creado este nuevo proceso, asi que cuando se invoque al detach
 *     con ese handler => el programa terminara la ejecucion.
 */
void Communicator::do_cloneChild()
throw() {
  LOGMETHOD(TraceMethod traceMethod("comm::Communicator", "do_cloneChild", ANNA_FILE_LOCATION));
  LOGINFORMATION(
    string msg("comm::Communicator::do_cloneChild | MainHandler: ");
    msg += a_mainHandler->asString();
    Logger::information(msg, ANNA_FILE_LOCATION);
  );
  Handler* handler(NULL);
  vector <Handler*> toDetach;

  for(handler_iterator ii = handler_begin(); ii != handler_end(); ii ++) {                        // (1)
    handler = Communicator::handler(ii);

    if(handler == a_mainHandler)
      continue;

    switch(handler->getType()) {
    case Handler::Type::Custom:
    case Handler::Type::RemoteConnection:
    case Handler::Type::ClientSocket:
      a_poll->erase(handler->getfd());                                                         // (4)
      LOGDEBUG(
        string msg("comm::Communicator::do_cloneChild | ");
        msg += handler->asString();
        Logger::debug(msg, ANNA_FILE_LOCATION);
      );
      handler->clone();
      a_poll->insert(handler->getfd());
      break;
    default:
      toDetach.push_back(handler);                                                             // (3)
      break;
    }
  }

  for(vector <Handler*>::iterator ii = toDetach.begin(), maxii = toDetach.end(); ii != maxii; ii ++)
    detach(*ii);

  toDetach.clear();

  try {
    singlethreadedAccept();
  } catch(RuntimeException& ex) {
    ex.trace();
  }

  exit(0);                                                                                       // (5)
}

int Communicator::SortByFileDescriptor::value(const Handler* handler)
throw() {
  return handler->getfd();
}


