// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#include <anna/diameter.comm/OamModule.hpp>
#include <anna/diameter.comm/Engine.hpp>
#include <anna/diameter.comm/LocalServer.hpp>
#include <anna/diameter.comm/ServerSession.hpp>
#include <anna/diameter.comm/ServerSocket.hpp>
#include <anna/core/functions.hpp>
#include <anna/statistics/Engine.hpp>
#include <anna/diameter.comm/TimerManager.hpp>

#include <anna/core/tracing/Logger.hpp>
#include <anna/core/tracing/TraceMethod.hpp>
#include <anna/core/functions.hpp>
#include <anna/app/functions.hpp>
#include <anna/xml/Node.hpp>
#include <anna/comm/Communicator.hpp>
#include <anna/comm/Network.hpp>
#include <anna/comm/Host.hpp>
#include <anna/comm/ClientSocket.hpp>

// STL
#include <string>



using namespace anna::diameter::comm;

LocalServer::LocalServer() :
  //a_key(key),
  a_description(""),
  a_maxConnections(-1),
  a_currentConnections(0),
  a_allowedInactivityTime(ServerSession::DefaultAllowedInactivityTime),
  a_engine(NULL),
  a_serverSocket(NULL),
  a_category(0),
  a_lock(false),
  a_available(false),
  a_lastUsedResource(NULL) {}


void LocalServer::initializeStatisticResources() throw() {
  std::string accName = "local server '";
  accName += anna::functions::socketLiteralAsString(a_key.first, a_key.second);
  accName += "' on origin-realm '";
  accName += a_engine ? a_engine->getOriginRealmName() : "unknown"; // it should be known (createServer)
  accName += "' and origin-host '";
  accName += a_engine ? a_engine->getOriginHostName() : "unknown"; // it should be known (createServer)
  accName += "'";
  a_messageStatistics.initialize(accName);
}

void LocalServer::resetStatistics() throw() {
  a_messageStatistics.getAccumulator()->reset();
}

void LocalServer::updateProcessingTimeStatisticConcept(const double &value, const anna::diameter::CommandId &cid) throw() {
  a_messageStatistics.process(MessageStatistics::ConceptType::SentRequestProcessingTime, cid, value);
  LOGDEBUG(anna::Logger::debug(a_messageStatistics.getAccumulator()->asString(), ANNA_FILE_LOCATION));
}

void LocalServer::updateReceivedMessageSizeStatisticConcept(const double &value, const anna::diameter::CommandId &cid) throw() {
  a_messageStatistics.process(MessageStatistics::ConceptType::ReceivedMessageSize, cid, value);
  LOGDEBUG(anna::Logger::debug(a_messageStatistics.getAccumulator()->asString(), ANNA_FILE_LOCATION));
}

ServerSession* LocalServer::allocateServerSession() throw() { return a_serverSessionsRecycler.create(); }
void LocalServer::releaseServerSession(ServerSession *serverSession) throw() { a_serverSessionsRecycler.release(serverSession); }


LocalServer::serverSession_iterator LocalServer::serverSession_find(const serverSession_key &key) throw() {
  return a_serverSessions.find(key);
}


LocalServer::serverSession_key LocalServer::getServerSessionKey(const anna::comm::ClientSocket &clientSocket) const throw() {
  return (anna::functions::hash(clientSocket.getRemoteAccessPoint().getINetAddress().serialize().c_str()));
}


void LocalServer::availabilityLost() throw() {
  a_available = false;
  std::string socket = anna::functions::socketLiteralAsString(a_key.first, a_key.second);
  LOGDEBUG(
    std::string msg = "diameter::comm::LocalServer { Socket: ";
    msg += socket;
    msg += " } has lost its availability";
    anna::Logger::debug(msg, ANNA_FILE_LOCATION);
  );
  // OAM
  OamModule &oamModule = OamModule::instantiate();
  oamModule.activateAlarm(OamModule::Alarm::c_LostAvailabilityOverLocalServerDefinedAs__s__, socket.c_str());
  oamModule.count(OamModule::Counter::LostAvailabilityOverLocalServer);
  a_engine->availabilityLost(this);
  a_engine->refreshAvailabilityForLocalServers();
}


void LocalServer::availabilityRecovered() throw() {
  a_available = true;
  std::string socket = anna::functions::socketLiteralAsString(a_key.first, a_key.second);
  LOGDEBUG(
    std::string msg = "diameter::comm::LocalServer { Socket: ";
    msg += socket;
    msg += " } has recovered its availability";
    anna::Logger::debug(msg, ANNA_FILE_LOCATION);
  );
  // OAM
  OamModule &oamModule = OamModule::instantiate();
  oamModule.cancelAlarm(OamModule::Alarm::c_LostAvailabilityOverLocalServerDefinedAs__s__, socket.c_str());
  oamModule.count(OamModule::Counter::RecoveredAvailabilityOverLocalServer);
  a_engine->availabilityRecovered(this);
  a_engine->refreshAvailabilityForLocalServers();
}



bool LocalServer::refreshAvailability() throw() {
  // Here available
  if(a_available) {  // check not-bound state for all server-sessions:
//      bool isolate = true;
//
//      for (const_serverSession_iterator it = serverSession_begin(); it != serverSession_end(); it++)
//         if (serverSession(it)->getState() != ServerSession::State::Closed) { isolate = false; break; }
//
//      if (isolate) {
// El problema de lo anterior, es que cuando se acepta una conexion, aun no ha llegado el CER (receive). Un server session
// esta en estado "Bound" cuando llega dicho CER y consecuentemente envio un CEA. Nos basaremos en 'a_currentConnections':
    if(a_currentConnections == 0) {
      availabilityLost();
      return true;
    }

    return false;
  }

  // Here not available
//   for (const_serverSession_iterator it = serverSession_begin(); it != serverSession_end(); it++)
//      if (serverSession(it)->getState() == ServerSession::State::Bound) {
  if(a_currentConnections > 0) {  // really == 0
    availabilityRecovered();
    return true;
  }

  return false;
}


void LocalServer::enable(bool unlock) throw(anna::RuntimeException) {
  // Unlock ?
  if(unlock) a_lock = false;

  if(a_lock) return;

  if(a_serverSocket && a_serverSocket->isOpened()) return;  // communicator attach twice gets poll bad file descriptor and application stops !

  // Resolve local address:
  anna::comm::Network& network = anna::comm::Network::instantiate();
  // Little tricky:
  anna::comm::Host *host = network.resolve(a_key.first /* addr */);
  const anna::comm::Device *device = *(host->device_begin());
  anna::comm::INetAddress localAddress(device, a_key.second /* port */);
  // Create server socket and assign receiver factory
  a_serverSocket = new ServerSocket(localAddress, this);
  a_serverSocket->setCategory(a_category);
  attach();
}

void LocalServer::attach() throw() {
  try {
    // Attach to communicator
    anna::comm::Communicator * communicator = anna::app::functions::component <anna::comm::Communicator> (ANNA_FILE_LOCATION);
    communicator->attach((anna::comm::ServerSocket*)a_serverSocket); // invokes handler insert and then initialize -> server socket bind (*)
    // OAM
    OamModule &oamModule = OamModule::instantiate();
    oamModule.count(OamModule::Counter::ServerSocketsOpened);
  } catch(anna::RuntimeException& ex) {
    ex.trace(); // fails on (*) (i.e. Address already in use), within communicator attach
    attachPlanning();
    a_serverSocket->close();
  }
}

void LocalServer::attachPlanning() throw() {
  LOGMETHOD(anna::TraceMethod tttm("diameter::comm::LocalServer", "attachPlanning", ANNA_FILE_LOCATION));

  try {
    TimerManager::instantiate().createTimer(this);
  } catch(anna::RuntimeException& ex) {
    ex.trace();
    anna::Logger::error("CAPTURED EXCEPTION activating attachPlanning timer", ANNA_FILE_LOCATION);
  }
}


void LocalServer::disable(bool lock) throw(anna::RuntimeException) {
  // Permanent ?
  a_lock = lock;
  anna::comm::Communicator * communicator = anna::app::functions::component <anna::comm::Communicator> (ANNA_FILE_LOCATION);
  communicator->detach((anna::comm::ServerSocket*)a_serverSocket);
  //delete(a_serverSocket);
  // OAM
  OamModule &oamModule = OamModule::instantiate();
  oamModule.count(OamModule::Counter::ServerSocketsClosed);
}


void LocalServer::lostConnection() throw() {
  a_currentConnections--;
  enable();
}


void LocalServer::newConnection() throw(anna::RuntimeException) {
  a_currentConnections++;

  // Check capacity
  if(a_currentConnections == a_maxConnections) {
    LOGWARNING(anna::Logger::warning("The maximum number of connections allowed over diameter server socket have already been served", ANNA_FILE_LOCATION));
    disable();
  }

  // Inform local server (availability changes):
  refreshAvailability();

  // OAM
  OamModule &oamModule = OamModule::instantiate();
  oamModule.count(OamModule::Counter::CreatedConnectionForServerSession);
}



ServerSession *LocalServer::createServerSession(const anna::comm::ClientSocket &clientSocket) throw(anna::RuntimeException) {
  LOGMETHOD(anna::TraceMethod tttm("diameter::comm::LocalServer", "createServerSession", ANNA_FILE_LOCATION));
  ServerSession* result(NULL);
  // First erase deprecated ones:
  std::vector<const ServerSession*> deprecated_server_sessions;
  const ServerSession* ss;

  for(const_serverSession_iterator it = serverSession_begin(); it != serverSession_end(); it++) {
    ss = serverSession(it);

    if(ss->a_deprecated)
      deprecated_server_sessions.push_back(ss);
  }

  std::vector<const ServerSession*>::iterator dc_ncit;
  std::vector<const ServerSession*>::iterator dc_min(deprecated_server_sessions.begin());
  std::vector<const ServerSession*>::iterator dc_max(deprecated_server_sessions.end());
  serverSession_iterator ii;

  for(dc_ncit = dc_min; dc_ncit != dc_max; dc_ncit++) {
    ii = serverSession_find((*dc_ncit)->getSocketId());
    a_serverSessions.erase(ii);
  }

  // End erase deprecated server sessions

  if((result = allocateServerSession()) == NULL)
    throw anna::RuntimeException("diameter::comm::LocalServer::allocateServerSession returns NULL", ANNA_FILE_LOCATION);

  // Initialize:
  result->initialize(); // warning: recycler does not initialize its objects and at least...
  // Assignments (it could be done at allocate):
  serverSession_key key = getServerSessionKey(clientSocket);
  result->setAllowedInactivityTime(getAllowedInactivityTime());
  result->setClientSocket((anna::comm::ClientSocket*)(&clientSocket));
  result->a_parent = this;
  result->a_socketId = key; // de momento...
  result->initializeSequences(); // despues de asignar el LocalServer y el socketId (sequences are seed-based by mean exclusive hash)
  result->a_engine = a_engine;
  a_serverSessions.insert(serverSession_value_type(key, result));
  newConnection();
  a_deliveryIterator = serverSession_begin();
  return result;
}



void LocalServer::closeServerSession(ServerSession* serverSession)
throw(anna::RuntimeException) {
  if(serverSession == NULL)
    return;

  LOGDEBUG(
    std::string msg("diameter::comm::LocalServer::closeServerSession | ");
    msg += serverSession->asString();
//      msg += " | Destroy: ";
//      msg += (destroy ? "yes" : "no");
    anna::Logger::debug(msg, ANNA_FILE_LOCATION);
  );
  serverSession_iterator ii = serverSession_find(serverSession->getSocketId());

  if(ii == serverSession_end())
    return;

  // Remove origin-realm / origin-host for server session in delivery map
  // This is related to http://redmine.teslayout.com/issues/41
  a_engine->manageDrDhServerSession(serverSession, false /* desregister */);

  try {
    //serverSession->setState(ServerSession::State::Closing); NOT MANAGED WITH SERVER SESSIONS
    serverSession->unbind(true /* always forceDisconnect on server sessions ... */);
    releaseServerSession(serverSession);
  } catch(anna::RuntimeException& ex) {
    ex.trace();
  }

  //a_serverSessions.erase(ii); // IMPORTANTE: posible fuente de cores de este tipo, en relacion con ServerSession::finalize() => delete(this)
  //   #0  0x0000003ca1c2e26d in raise () from /lib64/tls/libc.so.6
  //   (gdb) bt
  //   #0  0x0000003ca1c2e26d in raise () from /lib64/tls/libc.so.6
  //   #1  0x0000003ca1c2fa6e in abort () from /lib64/tls/libc.so.6
  //   #2  0x0000003ca8cb1148 in __gnu_cxx::__verbose_terminate_handler () from /usr/lib64/libstdc++.so.6
  //   #3  0x0000003ca8caf176 in __cxa_call_unexpected () from /usr/lib64/libstdc++.so.6
  //   #4  0x0000003ca8caf1a3 in std::terminate () from /usr/lib64/libstdc++.so.6
  //   #5  0x0000003ca8caf1b6 in std::terminate () from /usr/lib64/libstdc++.so.6
  //   #6  0x0000003ca8caf0c8 in __cxa_call_unexpected () from /usr/lib64/libstdc++.so.6
  //   #7  0x000000000047a4a7 in anna::diameter::comm::LocalServer::lostConnection (this=0x8aeb10) at comm.db/diameter.comm.LocalServer.cc:200
  //   #8  0x000000000047a9e6 in anna::diameter::comm::LocalServer::closeServerSession (this=0x8aeb10, serverSession=0xc37a00)
  //       at comm.db/diameter.comm.LocalServer.cc:275
  //   #9  0x000000000048d288 in anna::diameter::comm::ServerSession::finalize (this=0xc37a00) at comm.db/diameter.comm.ServerSession.cc:510
  //   #10 0x0000000000494e4f in anna::diameter::comm::ServerSessionReceiver::eventBreakLocalConnection (this=0xc119c0, clientSocket=@0xb0ea00)
  // SOLUCION: no borrar aqui, marcar como "deprecated". Este estado no se necesita realmente puesto que nadie volvera a usar este recurso.
  // Pero simplemente se podria usar para purgar mediante temporizacion (entonces s� se har�a el erase)
  serverSession->a_deprecated = true;
  // WE WILL ERASE AT createServerSession
  a_deliveryIterator = serverSession_begin();
  lostConnection();
}


ServerSession* LocalServer::findServerSession(int socketId, anna::Exception::Mode::_v emode)
throw(anna::RuntimeException) {
  serverSession_iterator ii = serverSession_find(socketId);

  if(ii != serverSession_end())
    return serverSession(ii);

  if(emode != anna::Exception::Mode::Ignore) {
    std::string msg("diameter::comm::LocalServer::findServerSession | SocketId: ");
    msg += anna::functions::asString(socketId);
    msg += " | ServerSession not found";
    anna::RuntimeException ex(msg, ANNA_FILE_LOCATION);

    if(emode == anna::Exception::Mode::Throw)
      throw ex;

    ex.trace();
  }

  return NULL;
}

ServerSession* LocalServer::findServerSession(const anna::comm::ClientSocket &clientSocket, anna::Exception::Mode::_v emode)
throw(anna::RuntimeException) {
  return findServerSession(getServerSessionKey(clientSocket), emode);
}


int LocalServer::getOTARequests() const throw() {
  int result = 0;

  for(const_serverSession_iterator it = serverSession_begin(); it != serverSession_end(); it++)
    result += serverSession(it)->getOTARequests();

  return result;
}

void LocalServer::close() throw(anna::RuntimeException) {
  LOGMETHOD(anna::TraceMethod tttm("diameter::comm::LocalServer", "close", ANNA_FILE_LOCATION));
  // Close listener (permanently to avoid reopening when local connections are being deleted):
  disable(true /* lock */);

  for(serverSession_iterator it = serverSession_begin(); it != serverSession_end(); it++)
    closeServerSession(serverSession(it));
}

void LocalServer::setClassCodeTimeout(const ClassCode::_v v, const anna::Millisecond & millisecond) throw() {
  LOGMETHOD(anna::TraceMethod tttm("diameter::comm::LocalServer", "setClassCodeTimeout", ANNA_FILE_LOCATION));

  for(serverSession_iterator it = serverSession_begin(); it != serverSession_end(); it++) {
    try {
      serverSession(it)->setClassCodeTimeout(v, millisecond);
    } catch(anna::RuntimeException &ex) {
      ex.trace();
    }
  }
}

void LocalServer::setMaxConnections(int maxConnections) throw(anna::RuntimeException) {
  LOGMETHOD(anna::TraceMethod tttm("anna::diameter::comm::LocalServer", "setMaxConnections", ANNA_FILE_LOCATION));

  // Negative & initial
  if(maxConnections < 0) {
    LOGDEBUG(anna::Logger::debug("Provided negative value means no limit accepting connections over server socket. Opening listen port (if closed)...", ANNA_FILE_LOCATION));
    a_maxConnections = -1;
    enable();
    return;
  }

  // No changes
  if(maxConnections == a_maxConnections) {
    LOGDEBUG(anna::Logger::debug("Provided equal to current. Ignore operation", ANNA_FILE_LOCATION));
    return;
  }

  // No margin
  if(maxConnections < a_currentConnections) {
    std::string msg = "There are more current connections (";
    msg += anna::functions::entriesAsString(a_currentConnections);
    msg += ") than provided maximum (";
    msg += anna::functions::entriesAsString(maxConnections);
    msg += "). Command rejected (you should release connections before logical limitation)";
    throw anna::RuntimeException(msg, ANNA_FILE_LOCATION);
  }

  // Updating
  if(maxConnections > a_currentConnections) {
    LOGDEBUG(anna::Logger::debug("Increasing connection margin (new limit is greater than current connections). Opening listen port (if closed)...", ANNA_FILE_LOCATION));
    enable();
  } else { // maxConnections == a_currentConnections: listen port must be closed if it is opened
    LOGDEBUG(

      if(maxConnections == 0)
      anna::Logger::debug("Provided zero value means disabling diameter server", ANNA_FILE_LOCATION);
      anna::Logger::debug("Zeroing connections margin (new limit is equal to current connections). Closing listen port (if opened)...", ANNA_FILE_LOCATION);
    );

    disable();
  }

  // Trace
  LOGDEBUG(
    std::string msg("Updating max connections from ");
    msg += (a_maxConnections == -1) ? "'no limit'" : anna::functions::asString(a_maxConnections);
    msg += " to ";
    msg += (maxConnections == -1) ? "'no limit'" : anna::functions::asString(maxConnections);
    anna::Logger::debug(msg, ANNA_FILE_LOCATION);
  );
  // Assignment
  a_maxConnections = maxConnections;
}


bool LocalServer::send(const Message* message, int socketId) throw(anna::RuntimeException) {
  LOGMETHOD(anna::TraceMethod tttm("diameter::comm::LocalServer", "send", ANNA_FILE_LOCATION));

  if(!isAvailable()) {
    LOGWARNING(
      std::string msg = "The local server ";

    if(a_description != "") {
    msg += "'";
    msg += a_description;
    msg += "' ";
  }
  msg += "is currently unavailable (no server sessions to send the message)";
  anna::Logger::warning(msg, ANNA_FILE_LOCATION);
  );
    return false;
  }

  if(socketId != -1) {  // socket id provided
    // Send (it was a request because of key (socketId) != -1, we forward the answer):
    try {
      ServerSession * fixedServerSession = a_engine->findServerSession(socketId); // exception if not found
      fixedServerSession->send(message);
      return true;
    } catch(anna::RuntimeException &ex) {
      std::string msg = "Cannot deliver answer through a fixed server session (socket id ";
      msg += anna::functions::asString(socketId);
      msg += "). Perhaps it existed but not now. Ignore";
      anna::Logger::error(msg, ANNA_FILE_LOCATION);
      ex.trace();
      return false;
    }
  }

  // Socket is not provided: use readSocketId
  socketId = (a_currentConnections > 1) ? readSocketId(message) : -1; // optimization

  if(a_deliveryIterator == serverSession_end()) a_deliveryIterator = serverSession_begin();

  a_lastUsedResource = (*a_deliveryIterator).second;

  if(socketId != -1) {
    a_lastUsedResource = findServerSession(socketId); // exception if not found
  } else { // Round Robin delivery between client-sessions
    if(getCurrentConnections() != 1) {  // optimize
      // Next server-session:
      a_deliveryIterator++;
    }
  }

  // Send:
  try {
    a_lastUsedResource->send(message);
    return true; // no matter if response is NULL (answers, i.e.) or not.
  } catch(anna::RuntimeException &ex) {
    ex.trace();
  }

  // Here, sent has failed:
  // OAM
  OamModule &oamModule = OamModule::instantiate();
  std::string socket = anna::functions::socketLiteralAsString(getKey().first, getKey().second);
  oamModule.activateAlarm(OamModule::Alarm::UnableToDeliverDiameterMessageToClientFromLocalServer__s__, socket.c_str());
  oamModule.count(OamModule::Counter::UnableToDeliverToClient);
  return false;
}


bool LocalServer::broadcast(const Message* message) throw(anna::RuntimeException) {
  LOGMETHOD(anna::TraceMethod tttm("diameter::comm::LocalServer", "broadcast", ANNA_FILE_LOCATION));
  bool allok = true;

  for(serverSession_iterator it = serverSession_begin(); it != serverSession_end(); it++) {
    try {
      serverSession(it)->send(message);
    } catch(anna::RuntimeException &ex) {
      ex.trace();
      allok = false;
    }
  }

  return allok;
}

void LocalServer::eventPeerShutdown(const ServerSession* serverSession) throw() {
  LOGWARNING(
    std::string msg(serverSession->asString());
    msg += " | eventPeerShutdown";
    anna::Logger::warning(msg, ANNA_FILE_LOCATION);
  );
}

void LocalServer::eventRequestRetransmission(const ServerSession* serverSession, Message *request) throw() {
  LOGWARNING(
    std::string msg(serverSession->asString());
    msg += " | eventRequestRetransmission";
    anna::Logger::warning(msg, ANNA_FILE_LOCATION);
  );
}

std::string LocalServer::asString() const throw() {
  std::string result("diameter::comm::LocalServer { ");
  result += "Description: ";
  result += (a_description != "") ? a_description : "undefined";
  result += " | Available (any server session bound): ";
  result += a_available ? "yes" : "no";
  result += " | Max Connections: ";
  result += anna::functions::asString(a_maxConnections);
  result += " | Current Connections: ";
  result += anna::functions::asString(a_currentConnections);
  // Current connections ??
  result += " | Allowed inactivity time for server sessions: ";
  result += a_allowedInactivityTime.asString();
  result += " | Server socket: ";
  result += a_serverSocket ? a_serverSocket->asString() : "closed";
  result += anna::functions::asString(" | OTA requests: %d%s", getOTARequests(), idle() ? " (idle)" : "");
  result += " | Last Incoming Activity Time: ";
  result += a_lastIncomingActivityTime.asString();
  result += " | Last Outgoing Activity Time: ";
  result += a_lastOutgoingActivityTime.asString();
//   result += "\n";
//   result += a_messageStatistics.getAccumulator()->asString();
  // ServerSessions only in xml
  return result += " }";
}


anna::xml::Node* LocalServer::asXML(anna::xml::Node* parent) const throw() {
  anna::xml::Node* result = parent->createChild("diameter.LocalServer");
  result->createAttribute("Description", (a_description != "") ? a_description : "undefined");
  result->createAttribute("Available", a_available ? "yes" : "no");
  result->createAttribute("MaxConnections", a_maxConnections);
  result->createAttribute("CurrentConnections", a_currentConnections);
  // Current connections ??
  result->createAttribute("AllowedInactivityTimeForServerSessions", a_allowedInactivityTime.asString());

  if(a_serverSocket)
    a_serverSocket->asXML(result);
  else
    result->createAttribute("ServerSocket", "closed");

  result->createAttribute("OTArequests", anna::functions::asString("%d%s", getOTARequests(), idle() ? " (idle)" : ""));
  result->createAttribute("LastIncomingActivityTime", a_lastIncomingActivityTime.asString());
  result->createAttribute("LastOutgoingActivityTime", a_lastOutgoingActivityTime.asString());
  // Statistics
  anna::xml::Node* stats = result->createChild("MessageStatistics");
  a_messageStatistics.getAccumulator()->asXML(stats);
  anna::xml::Node* serverSessions = result->createChild("ServerSessions"); // LocalServer.ServerSessions

  for(const_serverSession_iterator it = serverSession_begin(); it != serverSession_end(); it++)
    serverSession(it)->asXML(serverSessions);

  return result;
}

//------------------------------------------------------------------------------
//------------------------------------ LocalServer::updateIncomingActivityTime()
//------------------------------------------------------------------------------
void LocalServer::updateIncomingActivityTime() throw() {
  a_lastIncomingActivityTime = anna::functions::millisecond();
  LOGDEBUG
  (
    std::string msg = "Updated INCOMING activity on local server (milliseconds unix): ";
    msg += anna::functions::asString(a_lastIncomingActivityTime.getValue());
    anna::Logger::debug(msg, ANNA_FILE_LOCATION);
  );
}


//------------------------------------------------------------------------------
//------------------------------------ LocalServer::updateOutgoingActivityTime()
//------------------------------------------------------------------------------
void LocalServer::updateOutgoingActivityTime(void) throw() {
  a_lastOutgoingActivityTime = anna::functions::millisecond();
  LOGDEBUG
  (
    std::string msg = "Updated OUTGOING activity on local server (milliseconds unix): ";
    msg += anna::functions::asString(a_lastOutgoingActivityTime.getValue());
    anna::Logger::debug(msg, ANNA_FILE_LOCATION);
  );
}
