// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


// Local
#include <anna/diameter/helpers/defines.hpp>
#include <anna/diameter/helpers/base/defines.hpp>
#include <anna/diameter/helpers/base/functions.hpp>
#include <anna/diameter/codec/Message.hpp>
#include <anna/diameter/codec/functions.hpp>
#include <anna/config/defines.hpp> // general types, decoding helpers (DECODE[2/3/4]BYTES_INDX_VALUETYPE), etc.

#include <anna/core/functions.hpp>
#include <anna/core/DataBlock.hpp>
#include <anna/core/util/Tokenizer.hpp>

// STL
#include <string>

using namespace anna;
using namespace anna::diameter::codec;
using namespace anna::diameter::helpers::base;

// getters
U32 anna::diameter::helpers::base::functions::getResultCode(const anna::DataBlock & db) throw(anna::RuntimeException) {
  if(db.getSize() < Message::HeaderLength)
    throw anna::RuntimeException("Not enough bytes to cover command header length", ANNA_FILE_LOCATION);

  //anna::DataBlock avpsDB(db.getData() + Message::HeaderLength, db.getSize() - Message::HeaderLength);
  //const char * resultCodePtr = anna::diameter::codec::functions::findAVP(avpsDB, AVPID__Result_Code);

  const char *avpsDB = db.getData() + Message::HeaderLength;
  int avpsLen = db.getSize() - Message::HeaderLength;
  const char * resultCodePtr = anna::diameter::codec::functions::findAVP(avpsDB, avpsLen, AVPID__Result_Code);

  if(resultCodePtr == NULL)
    throw anna::RuntimeException("Result-Code AVP not found in DataBlock provided", ANNA_FILE_LOCATION);

  // Decoded avp information:
  AvpId _id;
  char _flags;
  int _length;
  std::string _data;
  anna::diameter::codec::functions::decodeAVP(resultCodePtr, _id, _flags, _length, _data);
  // Result:
  U32 result = DECODE4BYTES_INDX_VALUETYPE(_data, 0, U32);
  return result;
}



void anna::diameter::helpers::base::functions::decodeSessionId(const std::string &sessionId,
    std::string &diameterIdentity, U32 &high, U32 &low, std::string &optional) throw(anna::RuntimeException) {
// RFC 3588:
//
//      8.8.  Session-Id AVP
//
//         The Session-Id AVP (AVP Code 263) is of type UTF8String and is used
//         to identify a specific session (see Section 8).  All messages
//         pertaining to a specific session MUST include only one Session-Id AVP
//         and the same value MUST be used throughout the life of a session.
//         When present, the Session-Id SHOULD appear immediately following the
//         Diameter Header (see Section 3).
//
//         The Session-Id MUST be globally and eternally unique, as it is meant
//         to uniquely identify a user session without reference to any other
//         information, and may be needed to correlate historical authentication
//         information with accounting information.  The Session-Id includes a
//         mandatory portion and an implementation-defined portion; a
//         recommended format for the implementation-defined portion is outlined
//         below.
//
//         The Session-Id MUST begin with the sender's identity encoded in the
//         DiameterIdentity type (see Section 4.4).  The remainder of the
//         Session-Id is delimited by a ";" character, and MAY be any sequence
//         that the client can guarantee to be eternally unique; however, the
//         following format is recommended, (square brackets [] indicate an
//         optional element):
//
//         <DiameterIdentity>;<high 32 bits>;<low 32 bits>[;<optional value>]
//
//         <high 32 bits> and <low 32 bits> are decimal representations of the
//         high and low 32 bits of a monotonically increasing 64-bit value.  The
//         64-bit value is rendered in two part to simplify formatting by 32-bit
//         processors.  At startup, the high 32 bits of the 64-bit value MAY be
//         initialized to the time, and the low 32 bits MAY be initialized to
//         zero.  This will for practical purposes eliminate the possibility of
//         overlapping Session-Ids after a reboot, assuming the reboot process
//         takes longer than a second.  Alternatively, an implementation MAY
//         keep track of the increasing value in non-volatile memory.
//
//         <optional value> is implementation specific but may include a modem's
//         device Id, a layer 2 address, timestamp, etc.
//
//         Example, in which there is no optional value:
//            accesspoint7.acme.com;1876543210;523
//
//         Example, in which there is an optional value:
//            accesspoint7.acme.com;1876543210;523;mobile@200.1.1.88
//
//         The Session-Id is created by the Diameter application initiating the
//         session, which in most cases is done by the client.  Note that a
//         Session-Id MAY be used for both the authorization and accounting
//         commands of a given application.
  anna::Tokenizer sid;
  sid.apply(sessionId, ";");

  if(sid.size() < 3)
    throw anna::RuntimeException("Session-Id AVP has wrong format: at least three semicolons ';' must be included", ANNA_FILE_LOCATION);

  anna::Tokenizer::const_iterator tok_min(sid.begin());
  anna::Tokenizer::const_iterator tok_max(sid.end());
  anna::Tokenizer::const_iterator tok_iter = tok_min;
  diameterIdentity = anna::Tokenizer::data(tok_iter); tok_iter++;
  std::string s_high = anna::Tokenizer::data(tok_iter); tok_iter++;
  std::string s_low = anna::Tokenizer::data(tok_iter); tok_iter++;
  optional = ((tok_iter != tok_max) ? (anna::Tokenizer::data(tok_iter)) : "");
  high = (U32)atoi(s_high.c_str());
  low = (U32)atoi(s_low.c_str());
}


std::string anna::diameter::helpers::base::functions::getSessionId(const anna::DataBlock & db,
    std::string &diameterIdentity, U32 &high, U32 &low, std::string &optional) throw(anna::RuntimeException) {
  if(db.getSize() < Message::HeaderLength)
    throw anna::RuntimeException("Not enough bytes to cover command header length", ANNA_FILE_LOCATION);

  //anna::DataBlock avpsDB(db.getData() + Message::HeaderLength, db.getSize() - Message::HeaderLength);
  //const char * sessionIdPtr = anna::diameter::codec::functions::findAVP(avpsDB, AVPID__Session_Id);
  const char *avpsDB = db.getData() + Message::HeaderLength;
  int avpsLen = db.getSize() - Message::HeaderLength;
  const char * sessionIdPtr = anna::diameter::codec::functions::findAVP(avpsDB, avpsLen, AVPID__Session_Id);


  if(sessionIdPtr == NULL)
    throw anna::RuntimeException("Session-Id AVP not found in DataBlock provided", ANNA_FILE_LOCATION);

  // Decoded avp information:
  AvpId _id;
  char _flags;
  int _length;
  std::string result;
  anna::diameter::codec::functions::decodeAVP(sessionIdPtr, _id, _flags, _length, result);
  // Aditional helper data:
  decodeSessionId(result.c_str(), diameterIdentity, high, low, optional);
  return result;
}


std::string anna::diameter::helpers::base::functions::getOriginHost(const anna::DataBlock & db) throw(anna::RuntimeException) {
  if(db.getSize() < Message::HeaderLength)
    throw anna::RuntimeException("Not enough bytes to cover command header length", ANNA_FILE_LOCATION);

  const char *avpsDB = db.getData() + Message::HeaderLength;
  int avpsLen = db.getSize() - Message::HeaderLength;
  const char * originHostPtr = anna::diameter::codec::functions::findAVP(avpsDB, avpsLen, AVPID__Origin_Host);


  if(originHostPtr == NULL)
    throw anna::RuntimeException("Origin-Host AVP not found in DataBlock provided", ANNA_FILE_LOCATION);

  // Decoded avp information:
  AvpId _id;
  char _flags;
  int _length;
  std::string result;
  anna::diameter::codec::functions::decodeAVP(originHostPtr, _id, _flags, _length, result);
  return result;
}

